/*******************************************************************************

   Entity is the base of every object in the game world. 
     
   Date: (2005): initial version
   
   Authors: Clay Smith
   
   Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
-------------------------------------------------------------------------------
import arc.gfx.entity;

int main() {
   
   // there is no point in using entity by itself, rather it is a base other 
   // game objects build themselves upon. Every game object gets entity's functions
   // and variables, therefore it is good to browse through them.

   return 0;
}
-------------------------------------------------------------------------------
*******************************************************************************/
module arc.gfx.entity; 


private import arc.io.window,
              arc.io.input,
              arc.io.camera,
              arc.phy.point,
              arc.io.serialization.serializer; 

private import 	derelict.opengl.gl,
				derelict.opengl.glu;

private import arc.phy.phyutil;

private import std.math;

// entity x and y is scheduled to be replaced with getX and getY, 
// same for width and height
        
/*******************************************************************************

   The entity class is the basis of all things in the game world

*******************************************************************************/
class Entity 
{
  public:

   this()
   {
   }

   ~this()
   {
   }

   /// initialize OpenGL parameters required to draw
   void initGL()
   {
      glEnable(GL_TEXTURE_2D); 
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   }

   /*******************************************************************************
   
      Test if mouse is over entity
   
   *******************************************************************************/
   bool mouseOver()
   {
      if (camera is null)
      {
         return xyCol(Point(arc.io.input.mouseX, arc.io.input.mouseY));
      }
      else
      {
         return xyCol(camera.getMousePos);
      }
   
      return false;
   }

   /*******************************************************************************
   
      Test if given coordinate is over entity, simple box collision
   
   *******************************************************************************/
   bool xyCol(Point p)
   {
			float max_size = fmax(size.getWidth, size.getHeight);

      if (camera is null) {
         if (p.getX > pos.getX && p.getX < pos.getX + max_size &&
             p.getY > pos.getY && p.getY < pos.getY + max_size) {
               return true;
         }
      }
      else
      {
         //BUG: The following is incorrect if the camera is rotated
         if (p.getX > pos.getX && p.getX < pos.getX + max_size / camera.getZoom &&
             p.getY > pos.getY && p.getY < pos.getY + max_size / camera.getZoom) {
               return true;
         }
      }
   
      return false;
   }

   /*******************************************************************************
   
      Left mouse clicks on entity
   
   *******************************************************************************/
   bool mouseHit()
   {
      if (mouseOver && arc.io.input.mouseHit(LEFT))
         return true; 

      return false; 
   }

   /// returns true if entity is on the screen
   bool onScreen() {
			float max_size = fmax(size.getWidth, size.getHeight);

      if( camera is null )
      {
         if( pos.getX >= -max_size && pos.getX <= arc.io.window.getWidth + max_size &&
             pos.getY >= -max_size && pos.getY <= arc.io.window.getHeight + max_size) 
            return true;
      }
      else
      {
         //BUG: Doesn't work with rotation.
         max_size /= camera.getZoom;
         if( pos.getX >= camera.getViewPos.getX - max_size && 
             pos.getX <= camera.getViewPos.getX + camera.getViewSize.getX + max_size &&
             pos.getY >= camera.getViewPos.getY - max_size &&
             pos.getY <= camera.getViewPos.getY + camera.getViewSize.getY + max_size)
            return true;
      }

      return false;
   }

/+ map? isn't that game specific?
   /*******************************************************************************
   
      Returns true if entity is on the screen, transitional coords are for 
         when the map isn't big enough
   
   *******************************************************************************/
   bool onScreen(float transV, float transH)
   {
      if (camera == null)
         if (pos.getX >= -transV - size.getWidth && pos.getX <= arc.io.window.getWidth - transV)
            if (pos.getY >= -transH - size.getHeight  && pos.getY <= arc.io.window.getHeight  /* - game.hud.getHeight*/ - transH)
               return true;
      else
      {
         //BUG: Screen is possibly rotated. Basically need (rotated box)-box collision here.
         //     is it really worth it? graphics adapters are pretty good at eliminating things
         //     that are not on screen anyway.
         return true;
      }
   
      return false;
   }

   /*******************************************************************************
   
      Returns true if entity is on the map
   
   *******************************************************************************/
   bool onMap(float mapWidth, float mapHeight)
   {
      if (pos.getX >= 0 && pos.getX <= mapWidth)   
         if (pos.getY >= 0 && pos.getY <= mapHeight /*- game.hud.getHeight*/)
            return true;
   
      return false;
   }
+/

/+ Do you need this?
   /*******************************************************************************
   
      Return true if X value is on the screen given tranlational coord
   
   *******************************************************************************/
   bool XonScreen(float transV)
   {
      if (pos.getX >= -transV - size.getWidth && pos.getX <= arc.io.window.getWidth - transV)
         return true;
   
      return false;
   }

   /*******************************************************************************
   
      Return true if Y value is on the screen given tranlational coord
   
   *******************************************************************************/
   bool YonScreen(float transH)
   {
      if (pos.getY >= -transH - size.getHeight  && pos.getY <= arc.io.window.getHeight /*- game.hud.getHeight*/ - transH)
         return true;
   
      return false;
   }

   /*******************************************************************************
   
      Return true if X center is on the screen given tranlational coord
   
   *******************************************************************************/
   bool XonScreenC(float transV)
   {
      if (pos.getX >= -transV - size.getWidth && pos.getX <= arc.io.window.getWidth/2 - transV)
         return true;
   
      return false;
   }

   /*******************************************************************************
   
      Return true if Y center is on the screen given tranlational coord

   *******************************************************************************/
   bool YonScreenC(float transH)
   {
      if (pos.getY >= -transH - size.getHeight  && pos.getY <= arc.io.window.getHeight/2 /*- game.hud.getHeight*/ - transH)
         return true;
   
      return false;
   }
+/

   /*******************************************************************************
   
      Be able to save/load information using the serializer
   
   *******************************************************************************/
   void describe(T)(T s)
   {
      s.describe(pos); 
      s.describe(size); 
   }

   /// set position
   void setPosition(float x, float y) { pos.setX(x); pos.setY(y); }
   /// set size
   void setSize(float width, float height) { size.setX(width); size.setY(height); }
   /// just set X value
   void setX(float x) { pos.setX(x); }
   /// just set Y value
   void setY(float y) { pos.setY(y); }
   /// add value to X
   void addX(float x) { pos.setX(pos.getX + x); }
   /// add value to Y
   void addY(float y) { pos.setY(pos.getY + y); }

	/// change color of entity
	void setColor(short aR, short aG, short aB, short aA) 
	{ 
		eR = aR; eB = aB; eG = aG; eA = aA;
	}

   /// returns red
   short getR() { return eR; }
   /// returns green
   short getG() { return eG; }
   /// returns blue
   short getB() { return eB; }
   /// returns alpha
   short getA() { return eA; }

   /// return x position
   float getX() { return pos.getX; }
   /// return y position
   float getY() { return pos.getY; }
   /// return entity width
   float getWidth() { return size.getWidth; }
   /// return entity height
   float getHeight() { return size.getHeight; }
   /// return center X position of entity
   float getCenterX() { return pos.getX + (size.getWidth/2); }
   /// return center Y position of entity
   float getCenterY() { return pos.getY + (size.getHeight/2); }
   /// return half of entity's width
   float getHalfWidth() { return getWidth/2; }
   /// return half of entity's height
   float getHalfHeight() { return getHeight/2; }
   /// return position as a point
   Point getPosition() { return pos; }
   /// return size as a point
   Point getSize() { return size; }

   /// returns true if entity is within proximity of x and y value
   bool inRange(real argX, real argY, real argProx)
   {
      // if x and y are within proximity distance
      if (arc.phy.phyutil.withinRange(getX, argX, argProx) && arc.phy.phyutil.withinRange(getY, argY, argProx))
      {  
         return true; // we are within range
      }
   
      return false; 
   }

	/// set width
	void setWidth(float argW)
	{
		size.setX(argW);
	}

	/// set height
	void setHeight(float argH)
	{
		size.setY(argH);
	}

   /***
      return the distance between two entities
   */
   real distance(inout Entity a)
   {
      return arc.phy.phyutil.distance(getCenterX,getCenterY,a.getCenterX,a.getCenterY);
   }

   /***
      return the distance between entity and two points
   */
   real distance(real x2, real y2)
   {
      return arc.phy.phyutil.distance(getCenterX,getCenterY,x2,y2);
   }
	 
   /***
      associate a camera with this entity
   */
   void setCamera(Camera c)
   {
      camera = c;
   }
		 

  private:
	// Everything in the game world needs a position, size, and color
	Point pos;
	Point size;
	
	// Everything is displayed through a camera or mode2D (in case camera is null)
  Camera camera = null;

	// color
	short eR=255, eG=255, eB=255, eA=255; 
}