/*******************************************************************************

	Freetype fonts for everyone!
     
	Date: (2006): 	Based on Nehe Freetype code, re-coded to the
					D language and to work specifically in Arc by
					Clay Smith.
   
   Authors: Clay Smith
   
   Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
-------------------------------------------------------------------------------
import arc.gfx.font;

int main() {
  
	Font font = new Font("test.ttf", 32);
	font.draw("text", x, y, r, g, b, a);
  
	return 0;
}
-------------------------------------------------------------------------------
*******************************************************************************/

module arc.gfx.font;

private import 	derelict.opengl.gl,
				derelict.opengl.glu,
				derelict.freetype.ft;
				
private import 	arc.phy.phyutil,
				arc.phy.staticcollision,
				arc.io.ioutil, 
				arc.gfx.texture;
				
private import 	std.string,
				std.utf; 
				
debug import std.stdio;

/// class used to construct freetype fonts for OpenGL
class Font 
{
  private:
	float h;			///< Holds the height of the font.
	float w;			///< Holds the width of the font.
	GLuint[] textures;	///< Holds the texture id's 
	GLuint list_base;	///< Holds the first display list id
	int numLines;       /// number of lines in font
	float maxWidth = 0; /// maximum width in font

  public:
	/// init based on freetype font file and given height
	this(char[] fname, uint argH) 
	{
		//Allocate some memory to store the texture ids.
		textures = new GLuint[128];

		h=argH;

		//Create and initilize a freetype font library.
		FT_Library library;
		if (FT_Init_FreeType( &library )) 
			throw new Exception("FT_Init_FreeType failed");

		//The object in which Freetype holds information on a given
		//font is called a "face".
		FT_Face face;

		//This is where we load in the font information from the file.
		//Of all the places where the code might die, this is the most likely,
		//as FT_New_Face will die if the font file does not exist or is somehow broken.
		if (FT_New_Face( library, toStringz(fname), 0, &face )) 
			throw new Exception("FT_New_Face failed (there is probably a problem with your font file)");

		//For some twisted reason, Freetype measures font size
		//in terms of 1/64ths of pixels.  Thus, to make a font
		//h pixels high, we need to request a size of h*64.
		//(h << 6 is just a prettier way of writting h*64)
		FT_Set_Char_Size( face, cast(int)h << 6, cast(int)h << 6, 96, 96);

		//Here we ask opengl to allocate resources for
		//all the textures and displays lists which we
		//are about to create.  
		list_base=glGenLists(128);

		// allocate room for our texture loader, so textures loaded 
		// after do not overwrite
		arc.gfx.texture.incrementTextureCount(128); 
		
		glGenTextures( 128, textures.ptr );

		//This is where we actually create each of the fonts display lists.		
		for(ubyte i=0;i < 128;i++)
			make_dlist(face,i);

		//We don't need the face information now that the display
		//lists have been created, so we free the assosiated resources.
		FT_Done_Face(face);

		//Ditto for the library.
		FT_Done_FreeType(library);
	}

	/// print auto-formatted string to the screen
	void draw(float x, float y, short r, short g, short b, short a, ... ) {
		dchar[] dtext = formatString(_arguments, _argptr); 
		char[] text = toUTF8(dtext);
		draw(text, x, y, r, g, b, a);
	}

	/// print freetype to the screen
	void draw(char[] text, float x, float y, short r, short g, short b, short a) {
		glColor4ub(r,g,b,a);
		
		// We want a coordinate system where things coresponding to window pixels.
		glPushAttrib(GL_TRANSFORM_BIT);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glPopAttrib();		
	
		GLuint font=list_base;
		float mh=h/.63f;	//We make the height about 1.5* that of
	
		glPushAttrib(GL_LIST_BIT | GL_CURRENT_BIT  | GL_ENABLE_BIT | GL_TRANSFORM_BIT);	
		glMatrixMode(GL_MODELVIEW);
		glDisable(GL_LIGHTING);
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	

		glListBase(font);

		float modelview_matrix[16];	
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview_matrix.ptr);

		char[][] lines = splitlines(text); 
		numLines = lines.length;

		//This is where the text display actually happens.
		//For each line of text we reset the modelview matrix
		//so that the line's text will start in the correct position.
		//Notice that we need to reset the matrix, rather than just translating
		//down by h. This is because when each character is
		//draw it modifies the current matrix so that the next character
		//will be drawn immediatly after it.  
		foreach (int i, char[] line; lines)
		{
			glPushMatrix();
			glLoadIdentity();
			glTranslatef(x,y+((i+1)*h),0);
			glMultMatrixf(modelview_matrix.ptr);

			//  The commented out raster position stuff can be useful if you need to
			//  know the length of the text that you are creating.
			//  If you decide to use it make sure to also uncomment the glBitmap command
			//  in make_dlist().
			glRasterPos2f(0,0);
			glCallLists(line.length, GL_UNSIGNED_BYTE, toStringz(line));
			float rpos[4];
			glGetFloatv(GL_CURRENT_RASTER_POSITION ,rpos.ptr);
			w=-x+rpos[0];

			if (w > maxWidth)
				maxWidth = w;

			glPopMatrix();
		}


		glPopAttrib();		

		glPushAttrib(GL_TRANSFORM_BIT);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glPopAttrib();
	}

	float getRealHeight() { return h; }

	float getTextWidth(char[] text)
	{
		// We want a coordinate system where things coresponding to window pixels.
		float calcWidth = 0, maxWidth=0;
		GLuint font=list_base;
		float mh=h/.63f;	//We make the height about 1.5* that of
		glListBase(font);

		float modelview_matrix[16];	
		glGetFloatv(GL_MODELVIEW_MATRIX, modelview_matrix.ptr);

		char[][] lines = splitlines(text);
		numLines = lines.length; 
		
		//This is where the text display actually happens.
		//For each line of text we reset the modelview matrix
		//so that the line's text will start in the correct position.
		//Notice that we need to reset the matrix, rather than just translating
		//down by h. This is because when each character is
		//draw it modifies the current matrix so that the next character
		//will be drawn immediatly after it.  
		foreach (int i, char[] line; lines)
		{
			
			glPushMatrix();
			glLoadIdentity();
			glColor4ub(0,0,0,0); 
			glTranslatef(0,0,0);
			glMultMatrixf(modelview_matrix.ptr);

			//  The commented out raster position stuff can be useful if you need to
			//  know the length of the text that you are creating.
			//  If you decide to use it make sure to also uncomment the glBitmap command
			//  in make_dlist().
			glRasterPos2f(0,0);
			glCallLists(line.length, GL_UNSIGNED_BYTE, toStringz(line));
			float rpos[4];
			glGetFloatv(GL_CURRENT_RASTER_POSITION ,rpos.ptr);
			calcWidth=rpos[0];

			if (calcWidth > maxWidth)
				maxWidth = calcWidth;

			glPopMatrix();
		}

		return maxWidth;
	}


	/// calculate index of freetype font
	int calculateIndex(char[] text, float x, float y, float mouseX, float mouseY) 
	{
		int index = -1; 

		char[][] lines = splitlines(text); 

		int i = 0;
		int prevLineCount = 0;
		foreach(char[] line; lines)
		{
			i++;

			// calc width and height
			float calcH = h*i;
			
			//writefln("calulated height is ", calcH);

			// if mouse is on current line
			if (mouseY <= calcH+y && mouseY >= calcH-h+y)
			{
				//writefln("at line ", line);
				index = searchIndex(line, cast(int)mouseX, cast(int)x, 0, text.length);
				//writefln("ret2 ", index+prevLineCount);
				return index + prevLineCount; 
			}

			prevLineCount+=line.length+1;
		}

		//writefln("return 0");
		return 0;
	}

	int searchIndex(char[] text, int mouseX, int posX, int left, int right)
	{
		//writefln(text.length, " is length"); 
		for (int j = 1; j <= text.length; j++)
		{
			//writefln("from 0 to ", j); 
			float a = getTextWidth(text[0 .. j])+posX;

			//writefln(a, " compared to ", mouseX, " of text ", text[0 .. j], " letter is ", text[j-1]);

			if (a > mouseX)
			{
				if (j > 0)
				{
				//	writefln("return ", j-1); 
					return j-1;
				}
				else
				{
					//writefln("return 0");
					return 0;
				}
			} 
		}

	//	writefln("return ", text.length);
		return text.length;
		//assert(0);
	}

	/// get height
	float getHeight() { return h*numLines; }
	/// get width
	float getWidth() { return maxWidth; }

	// Free all the resources assosiated with the font.
	~this() 
	{
		glDeleteLists(list_base,128);
		glDeleteTextures(128,textures.ptr);
		delete textures;
	}



  private:
 
	///Create a display list coresponding to the give character.
	void make_dlist ( FT_Face face, char ch) 
	{
		//The first thing we do is get FreeType to render our character
		//into a bitmap.  This actually requires a couple of FreeType commands:

		//Load the Glyph for our character.
		if(FT_Load_Glyph( face, FT_Get_Char_Index( face, ch ), FT_LOAD_DEFAULT ))
			throw new Exception("FT_Load_Glyph failed");

		//Move the face's glyph into a Glyph object.
		FT_Glyph glyph;
		if(FT_Get_Glyph( face.glyph, &glyph ))
			throw new Exception("FT_Get_Glyph failed");

		//Convert the glyph to a bitmap.
		FT_Glyph_To_Bitmap( &glyph, FT_Render_Mode.FT_RENDER_MODE_NORMAL, cast(FT_Vector*)0, 1 );
		FT_BitmapGlyph bitmap_glyph = cast(FT_BitmapGlyph)glyph;

		//This reference will make accessing the bitmap easier
	

		//Use our helper function to get the widths of
		//the bitmap data that we will need in order to create
		//our texture.
		int width = nextPowerOf2( bitmap_glyph.bitmap.width );
		int height = nextPowerOf2( bitmap_glyph.bitmap.rows );

		//Allocate memory for the texture data.
		GLubyte[] expanded_data = new GLubyte[ 2 * width * height];

		//Here we fill in the data for the expanded bitmap.
		//Notice that we are using two channel bitmap (one for
		//luminocity and one for alpha), but we assign
		//both luminocity and alpha to the value that we
		//find in the FreeType bitmap. 
		//We use the ?: operator so that value which we use
		//will be 0 if we are in the padding zone, and whatever
		//is the the Freetype bitmap otherwise.
		for(int j=0; j <height;j++) {
			for(int i=0; i < width; i++){
				expanded_data[2*(i+j*width)]= expanded_data[2*(i+j*width)+1] = 
					(i>=bitmap_glyph.bitmap.width || j>=bitmap_glyph.bitmap.rows) ?
					0 : bitmap_glyph.bitmap.buffer[i + bitmap_glyph.bitmap.width*j];
			}
		}


		//Now we just setup some texture paramaters.
		glBindTexture( GL_TEXTURE_2D, textures[ch]);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

		//Here we actually create the texture itself, notice
		//that we are using GL_LUMINANCE_ALPHA to indicate that
		//we are using 2 channel data.
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height,
			0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expanded_data.ptr );

		//With the texture created, we don't need to expanded data anymore
		delete expanded_data;

		//So now we can create the display list
		glNewList(list_base+ch,GL_COMPILE);

		glBindTexture(GL_TEXTURE_2D,textures[ch]);



		//Now we move down a little in the case that the
		//bitmap extends past the bottom of the line 
		//(this is only true for characters like 'g' or 'y'.
		glPushMatrix();

		//first we need to move over a little so that
		//the character has the right amount of space
		//between it and the one before it.
		//glTranslatef(bitmap_glyph.left,0,0);
		glTranslatef(bitmap_glyph.left,-bitmap_glyph.bitmap.rows-(bitmap_glyph.top-bitmap_glyph.bitmap.rows),0);

		//Now we need to account for the fact that many of
		//our textures are filled with empty padding space.
		//We figure what portion of the texture is used by 
		//the actual character and store that information in 
		//the x and y variables, then when we draw the
		//quad, we will only reference the parts of the texture
		//that we contain the character itself.
		float	x=cast(float)bitmap_glyph.bitmap.width / cast(float)width,
				y=cast(float)bitmap_glyph.bitmap.rows / cast(float)height;

		//Here we draw the texturemaped quads.
		//The bitmap that we got from FreeType was not 
		//oriented quite like we would like it to be,
		//so we need to link the texture to the quad
		//so that the result will be properly aligned.
		glBegin(GL_QUADS);
		glTexCoord2d(0,0); glVertex2f(0,0);//glVertex2f(0,bitmap_glyph.bitmap.rows);
		glTexCoord2d(0,y); glVertex2f(0,bitmap_glyph.bitmap.rows);//glVertex2f(0,0);
		glTexCoord2d(x,y); glVertex2f(	bitmap_glyph.bitmap.width, bitmap_glyph.bitmap.rows);//glVertex2f(bitmap_glyph.bitmap.width,0);
		glTexCoord2d(x,0); glVertex2f(bitmap_glyph.bitmap.width,0);//glVertex2f(	bitmap_glyph.bitmap.width, bitmap_glyph.bitmap.rows);
		glEnd();
		glPopMatrix();
		glTranslatef(face.glyph.advance.x >> 6 ,0,0);

		//increment the raster position as if we were a bitmap font.
		//(only needed if you want to calculate text length)
		glBitmap(0,0,0,0,face.glyph.advance.x >> 6,0,null);

		//Finnish the display list
		glEndList();
	}

}



