/*******************************************************************************

   Particle engine 
   
   Date:   2005 
    
   Authors: Original code was created by Jeff Molofee 1999 <BR>
            ported to Linux/SDL by Ti Leggett 2001<BR>
            ported to D/Arc and heavily modified by Clay Smith 2005
   
   Liscense: 
      See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
-------------------------------------------------------------------------------
import arc.gfx.particle;

int main() {
   

   return 0;
}
-------------------------------------------------------------------------------
*******************************************************************************/

module arc.gfx.particle;

// openGL + SDL
private import derelict.opengl.gl,
               derelict.sdl.sdl,  
               derelict.sdl.image; 

// ARC
private import arc.gfx.texture, 
               arc.gfx.entity;

private import arc.phy.phyutil;

private import std.c.stdlib;

// std
private import std.string,
				std.stdio,
				std.utf; 

// A single particle
struct Particle
{
   // Alive (Yes/No)
   int   alive; 

   // Particle Life
   short life;   

   // Fade Speed
   float fade;   

   // color   
   short r, g, b;     

   // getX and getY position
   float x, y;

   // getX and getY direction   
   float xi, yi;    
   
   // getX and getY gravity
   float xg, yg;     
} 

// make array assigning easy
void assign3arr(short[3] src, short[3] dst)
{
   for (int i = 0; i < 3; i++)
      src[i] = dst[i];
}

/// A group of particles
class Particles : Entity
{
  public:
	/// load particles based on user given vars
   this(char[] argFileName, 
		int argNum, int argSlow, int argXSpeedMin, int argXSpeedMax,
		int argYSpeedMin, int argYSpeedMax, int argXGrav, int argYGrav,
		short[3][] argColors)
		
	{
		debug writefln("particle: this()");

		numParticles = argNum;
		slowdown = argSlow;

		xSpeedMin = argXSpeedMin; 
		xSpeedMax = argXSpeedMax;

		ySpeedMin = argYSpeedMin;
		ySpeedMax = argYSpeedMax;

		xgrav = argXGrav;
		ygrav = argYGrav; 
   
		// set dyn arr length as appropriate
		parts.length = numParticles;  
		colors.length = argColors.length;

		// copy gColors into our own colors
		for (int i = 0; i < argColors.length; i++)
		{
			assign3arr(colors[i], argColors[i]);
			//debug writefln(colors[i][0], ",", colors[i][1], ",", colors[i][2]);
		}

		// load up particle image
		auto Texture tex = arc.gfx.texture.texture(argFileName, 1.0);

		texture = tex.ID; 

		// set size to size of img
		setSize(tex.getWidth, tex.getHeight);

		// initialize all particles
		for (uint loop = 0; loop < numParticles; loop++ )
		{
			reset(loop);  
		}
   }

   ~this()
   {
   }

   /// creates an explosion with added speed
   void explode(int argSpeed)
   {
      // add to speeds for quick explosion
      xSpeedMin -= argSpeed; 
      xSpeedMax += argSpeed; 
      ySpeedMin -= argSpeed;  
      ySpeedMax += argSpeed;

      for (uint loop = 0; loop < numParticles; loop++ )
      {
         reset(loop);  
      }

      // set back to normal
      xSpeedMin += argSpeed; 
      xSpeedMax -= argSpeed; 
      ySpeedMin += argSpeed;  
      ySpeedMax -= argSpeed;
   }

   /// draw particles and update to next position
   void draw(double gameSpeed)
   {
      // Modify each of the particles  
      for (uint loop = 0; loop < numParticles; loop++ )
      {
         if ( parts[loop].alive )
         {
            // Grab Our Particle X Position  
            float ex = parts[loop].x;

            // Grab Our Particle Y Position  
            float why = parts[loop].y;
   
            // only fade if ready to 
            short alpha = parts[loop].life;
            if (alpha > 255)
               alpha = 255;

            // Draw The Particle Using Our RGB Values,
            // Fade The Particle Based On It's Life
            glColor4ub(parts[loop].r,
                       parts[loop].g,
                       parts[loop].b,
                       alpha);
   
            // bind our texture to this opengl triangle strip
            glBindTexture( GL_TEXTURE_2D, texture);
   
            // Build Quad From A Triangle Strip  
            glBegin( GL_TRIANGLE_STRIP );
               // Top Right  
               glTexCoord2d( 1, 1 );
               glVertex2f( ex + getWidth, why + getHeight);
               // Top Left  
               glTexCoord2d( 0, 1 );
               glVertex2f( ex - getWidth, why + getHeight);
               // Bottom Right  
               glTexCoord2d( 1, 0 );
               glVertex2f( ex + getWidth, why - getHeight);
               // Bottom Left  
               glTexCoord2d( 0, 0 );
               glVertex2f( ex - getWidth, why - getHeight);
            glEnd( );
   
            // Move On The X Axis By X Speed  
            parts[loop].x += parts[loop].xi * gameSpeed / ( slowdown * 1000);
            // Move On The Y Axis By Y Speed  
            parts[loop].y += parts[loop].yi * gameSpeed / ( slowdown * 1000);
   
            // Take Pull On X Axis Into Account  
            parts[loop].xi += parts[loop].xg * gameSpeed;
            // Take Pull On Y Axis Into Account  
            parts[loop].yi += parts[loop].yg * gameSpeed;
   
            // Reduce Particles Life By 'Fade'  
            parts[loop].life -= (parts[loop].fade * gameSpeed) * 3;
   
            // If the particle dies, revive it  
            if ( parts[loop].life < 0.0f )
            {
               reset(loop);
            }
         }
      }
   }
   
  private:
   // reset a given particle to given initial state
   void reset(int num)
   {
      // Make the particels active  
      parts[num].alive = true;
      // Give the particles life  
      parts[num].life = random_range(255, 255*10);
      // Random Fade Speed  
      parts[num].fade = 1; 

      // pick random color in our color range
      int i = random_range(0,colors.length-1); 

      // set red   
      parts[num].r = colors[i][0];
      // Select Green Rainbow Color  
      parts[num].g = colors[i][1];
      // Select Blue Rainbow Color  
      parts[num].b = colors[i][2];
      // Set the position on the X axis  
      parts[num].x = getX; 
      parts[num].y = getY; 

      // Random Speed On X Axis  
      parts[num].xi = random_range(xSpeedMin, xSpeedMax);
      // Random Speed On Y Axi  
      parts[num].yi = random_range(ySpeedMin, ySpeedMax);
      // Set Horizontal Pull To Zero  
      parts[num].xg = xgrav;
      // Set Vertical Pull Downward  
      parts[num].yg = ygrav;
   }

  private:
   int numParticles;
   float slowdown;
   int xSpeedMin, xSpeedMax, ySpeedMin, ySpeedMax;
   float xgrav, ygrav;
   uint delay, texture;
   Particle[] parts;
   short[3][] colors;
}

