﻿/*******************************************************************************

Description: Graphics primitives

Authors: Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
--------------------
import 	arc.io.window, 
		arc.io.input,
		arc.gfx.prim;

// main is just here to handle states
int main()
{
	// open window with specific settings
	arc.io.window.open("Arc Graphics Primitives", 640, 480, false);

	// main loop
	while (!arc.io.input.keyDown(SDL_QUIT))
	{ 
		// clear the screen before drawing
		arc.io.window.clear();
		arc.io.input.process();

		drawPixel(arc.io.input.mouseX, arc.io.input.mouseY, 
					255, 0,0);

		drawCircle(arc.io.input.mouseX+100, arc.io.input.mouseY+10,
					25, 20, 255,0,0, false);

		drawLine(0,0,arc.io.input.mouseX, arc.io.input.mouseY, 0,0,255);

		drawRect(arc.io.input.mouseX, arc.io.input.mouseY+100,
				50, 100, 0,0,255,true);

		// flip to next screen
		arc.io.window.swap();
	}

	// close window when done with it
	arc.io.window.close();

	return 0;
}



--------------------

*******************************************************************************/

module arc.gfx.prim;

private import derelict.opengl.gl; 
private import arc.phy.point; 

debug import std.stdio;

/// draw pixel at position and color
void drawPixel(float x, float y, short r, short g, short b, short a)
{
	// disable gl textures
	glDisable(GL_TEXTURE_2D);

	// set color to one given
	glColor4ub(r, g, b, a); 

	// make sure the line width is only 1 pixel wide
	glLineWidth(1);

	// draw line
	glBegin(GL_LINE_LOOP);

	glVertex2f(x, y);
	glVertex2f(x+1, y); 

	glEnd();
}

/// draw line with color
void drawLine(	float x1, float y1, float x2, float y2, 
				short r, short g, short b, short a)
{
	// disable gl textures
	glDisable(GL_TEXTURE_2D);

	// set color to one given
	glColor4ub(r, g, b, a); 

	// draw line
	glBegin(GL_LINE_LOOP);

	glVertex2f(x1, y1);
	glVertex2f(x2, y2); 

	glEnd();
}

/// draw circle at position, size (radius), detail (vertex's), and color
void drawCircle(float x, float y, float radius, int detail, 
				short r, short g, short b, short a, bool fill)
{
	// primitives can only be drawn once textures are disabled
	glDisable(GL_TEXTURE_2D); 

	// set color to one given
	glColor4ub(r, g, b, a); 

	// we will be drawing lines
	if (fill)
	{
		glBegin(GL_POLYGON);
	}
	else
	{		
		glBegin(GL_LINE_LOOP);
	}

	Point p; 
 
	for (float i = 0; i < 360; i+= 360.0/detail)
	{
		// create polar coordinate
		p = Point(radius, i);
		// translate it to rectangular
		p.polarToRect();
		// and draw it
		glVertex2f(x+p.getX, y+p.getY);
      }
      
	glEnd(); 
}

/// draw rectange with given position, size, and color
void drawRect(	float x, float y, float width, float height, 
				short r, short g, short b, short a, bool fill)
{
	// disable images
	glDisable(GL_TEXTURE_2D); 

	// set color to one given
	glColor4ub(r, g, b, a); 
      
	// draw a box line of border
	if (fill)
	{
		glBegin(GL_POLYGON);
	}
	else
	{
		glBegin(GL_LINE_LOOP);
    }     

	// draw box to the screen
	glVertex2f(x+width, y); 
      
	glVertex2f(x+width, y+height);
      
	glVertex2f(x, y + height);
      
	glVertex2f(x, y);     
      
	glEnd(); 
}

/// simply draw image to screen with given image ID
void drawImage(	float x, float y, float width, float height,
				short r, short g, short b, short a, 
				float pivX, float pivY,
				float angle, int ID)
{
	// center calculations
	float halfWidth = width/2;
	float halfHeight = height/2;
	
	// enable 2d textures and bind texture
	glEnable(GL_TEXTURE_2D);
	
	// bind texture ID to this tex
	glBindTexture(GL_TEXTURE_2D, ID);  

	// set color to one given
	glColor4ub(r, g, b, a); 
      
	glPushMatrix();

	// rotate and translate
	glTranslatef(x,y,0);
	glRotatef(angle, 0, 0, 1);

	// draw image at given coords, binding texture appropriately   
	glBegin(GL_QUADS);
	  
	// -halfWidth and -halfHeight act as ancors to rotate from the center
	glTexCoord2d(0,0); glVertex2f(-halfWidth + pivX, -halfHeight + pivY); 
      
	glTexCoord2d(0,1); glVertex2f(	-halfWidth + pivX, 
									-halfHeight + height + pivY);
      
	glTexCoord2d(1,1); glVertex2f(	-halfWidth + width + pivX, 
									-halfHeight + height + pivY);
      
	glTexCoord2d(1,0); glVertex2f(	-halfWidth + width + pivX, 
									-halfHeight + pivY);     
      
	glEnd(); 

	glPopMatrix();
}


/// draw image from the top left location 
void drawImageTopLeft(float x, float y, float w, float h, ushort r, ushort b, ushort g, ushort a, int ID)
{
	// enable 2d textures and bind texture
	glEnable(GL_TEXTURE_2D);
	
	// bind texture ID to this tex
	glBindTexture(GL_TEXTURE_2D, ID);  

	// set color to one given
	glColor4ub(r,g,b,a);
      
	glPushMatrix();

	// rotate and translate
	glTranslatef(x,y,0);

	// draw image at given coords, binding texture appropriately   
	glBegin(GL_QUADS);
	  
	// -halfWidth and -halfHeight act as ancors to rotate from the center
	glTexCoord2d(0,0); glVertex2f(0 , 0); 
      
	glTexCoord2d(0,1); glVertex2f(0, h);
      
	glTexCoord2d(1,1); glVertex2f(w , h);
      
	glTexCoord2d(1,0); glVertex2f(w , 0);     
      
	glEnd(); 

	glPopMatrix();
}
