// sprite.d

/*

Possible Collisions

Box
	- Box
	- Circle
	- XY

Circle
	- Circle
	- XY

*/


module arc.gfx.sprite;

private import std.math; 

private import 	arc.gfx.entity,
				arc.gfx.prim,
				arc.gfx.texture,
				arc.gfx.frame.boxframe,
				arc.gfx.frame.radframe,
				arc.gfx.animation,
				arc.io.sound,
				arc.phy.staticcollision,
				arc.phy.phyutil,
				arc.phy.triangle,
				arc.phy.time,
				arc.phy.point;

// for COLLISION_TYPE
public import arc.gfx.frame.frame;

private import derelict.opengl.gl,
               derelict.sdl.sdl,
               derelict.sdl.image;

debug import arc.phy.phyutil;

import std.string,
       std.stdio;

const int DEFAULT_TIME = -1;
const char[] DEFAULT_ANIM = "default";

/// A graphical object with animation and sound
class Sprite : Entity
{
  public:	
	/// box/radius constructor, load from filename
	this(char[] filename, COLLISION_TYPE argCol)
	{
		// set collision type
		col = argCol;
		addFrame(filename, DEFAULT_ANIM, DEFAULT_TIME, null);
		initialize();
	}

	/// box/radius constructor, load from texture
	this(Texture tex, COLLISION_TYPE argCol)
	{
		// set collision type
		col = argCol;
		addFrame(tex, DEFAULT_ANIM, DEFAULT_TIME, null);
		initialize();
	}

	/// simple box/rad constructor, do not load anything
	this(COLLISION_TYPE argCol)
	{
		col = argCol;
		initialize();
	}

	// what everything needs to do to initialize itself
	void initialize()
	{
		zoomX = zoomY = 1;
		rotSpeed = new Blinker;
		rotPoint.length = 0;
	}

	// cleanup
	~this()  
	{
		// delete all animations
		foreach(char[] k; anim.keys)
			anim.remove(k);

		delete rotSpeed;
		rotPoint.length = 0;
	}

	/// add a frame to the animation
	void addFrame(char[] img, char[] animation, int time, Sound snd)
	{
		// allocate new animation as required
		if (!(animation in anim)) {
			anim[animation] = new Animation;
			anim.rehash;
		}
		// otherwise just add it and set curr animation to it
		currAnim = animation;
			
		// load frame and set size
		Point s = anim[animation].addFrame(img, time, snd, col, getGap);

		// set getWidth from one extracted from animation
		setSize(s.getWidth, s.getHeight);
		origW = s.getWidth;
		origH = s.getHeight;

		// calculate radius if necessary
		if (col == COLLISION_TYPE.RADIUS)
			calcRadius();
	}

	/// add a frame to the animation
	void addFrame(Texture tex, char[] animation, int time, Sound snd)
	{
		// allocate new animation as required
		if (!(animation in anim)) {
			anim[animation] = new Animation;
			anim.rehash;
		}
		// otherwise just add it and set curr animation to it
		currAnim = animation;
			
		// load frame and set size
		Point s = anim[animation].addFrame(tex, time, snd, col, getGap);

		// set getWidth from one extracted from animation
		setSize(s.getWidth, s.getHeight);
		origW = s.getWidth;
		origH = s.getHeight;

		// calculate radius if necessary
		if (col == COLLISION_TYPE.RADIUS)
			calcRadius();
	}

	/// returns true if point getX, getY collides with sprite
	bool collide(float argX, float argY)
	{
		if (col == COLLISION_TYPE.BOX)
		{
			return boxXYCollision(argX, argY, getLeftX, getTopY, getWidth, getHeight);
		}

		if (col == COLLISION_TYPE.RADIUS)
		{
			return circleXYCollision(argX, argY, getLeftX+getHalfWidth, getTopY+getHalfHeight, radius); 
		}

		arc.io.log.write("sprite", "collide(x,y)", "error", "no collision type found"); 
	}

	/// collide a sprite with any other sprite
	bool collide(Sprite s)
	{	
		if (col == COLLISION_TYPE.BOX)
		{
			// box-box
			if (s.col == COLLISION_TYPE.BOX)
			{
				return boxBoxCollision(	getLeftX, getTopY, getWidth, getHeight, s.getLeftX, s.getTopY, s.getWidth, s.getHeight);
			}
			// box-rad
			else if (s.col == COLLISION_TYPE.RADIUS)
			{
				return boxCircleCollision(	getLeftX, getTopY, getWidth, getHeight, 
											s.getLeftX+s.getHalfWidth, s.getTopY+s.getHalfHeight,
											s.getRadius);
			}
		}
		else if (col == COLLISION_TYPE.RADIUS)
		{
			// test circle
			if (s.col == COLLISION_TYPE.RADIUS)
			{
				return circleCircleCollision(getLeftX+getHalfWidth, getTopY+getHalfHeight,
											radius,
											s.getLeftX+s.getHalfWidth,
											s.getTopY+s.getHalfHeight,
											s.getRadius);
			}
			if (s.col == COLLISION_TYPE.BOX)
			{
				return boxCircleCollision(	s.getLeftX, s.getTopY, s.getWidth, s.getHeight, 
											getLeftX+getHalfWidth, getTopY+getHalfHeight,
											radius);
			}
		}

		arc.io.log.write("sprite", "collide(Sprite)", "error", "collision failed"); 
	}

	/// draw frame 
	void draw()
	{	
		anim[currAnim].draw(getX,getY,getWidth,getHeight,getR,getG,getB,getA,pivX,pivY,getAngle,getGap,getLeftX, getTopY,col);
	}

	/// draw the boundry of given type, won't draw anything for 'per-pixel'
	void drawBounds()
	{
		if (col == COLLISION_TYPE.BOX)
			drawRect(getLeftX, getTopY, getWidth, getHeight, getR,getG,getB,getA, false);
		else if (col == COLLISION_TYPE.RADIUS)
		{
			drawCircle(	getLeftX+getHalfWidth, getTopY+getHalfHeight, radius,
						30, getR,getG,getB,getA,false);
		}
	}

	/// fast angle setting
	void setFastAngle(float argA) {
		angle = argA;
	}

	/// rotate frame to specified degree
	void setAngle(float argA)
	{
		// rotate exactly to it
		if (rotateBy == 0)
		{
			angle = argA;		
			wrapAngle(angle);
		}
		// slowly rotate by setting target getAngle
		else
		{
			targetAngle = argA;		
			wrapAngle(targetAngle);
		}
	}
	
	/// get getAngle at which it is rotated at
	float getAngle() 
	{ 
		return angle; 
	}

	/// point on which to rotate around
	void setPivot(float argX, float argY)
	{
		pivX = argX; pivY = argY;
	}

	/// set X pivot point
	void setPivX(float argX)
	{
		pivX = argX;
	}

	/// set Y pivot point
	void setPivY(float argY)
	{
		pivY = argY;
	}

	/// returns X rotation pivot point
	float getPivX() { return pivX; }
	
	/// returns Y rotation pivot point
	float getPivY() { return pivY; }

	/// get radius
	float getRadius() 
	{ 
		return radius;
	}

	/// set radius
	void setRadius(float argR)
	{
		radius = argR;
	}

	/// return the current frame in the sprite
	Frame getCurrFrame()
	{
		return anim[currAnim].getCurrFrame;
	}

	/// set animation loop/not loop
	void setAnimLoop(char[] animation, bool v)
	{
		anim[animation].setLoop(v);
	}

	/// get zoom X value
	float getZoomX() { return zoomX; }

	/// get zoom Y value
	float getZoomY() { return zoomY; }

	/// set getWidth of sprite
	void setWidth(float argW)
	{
		setSize(argW, getHeight);
	}

	/// set getHeight of sprite
	void setHeight(float argH)
	{
		setSize(getWidth, argH);
	}

	/// set size of sprite
	void setSize(float argW, float argH)
	{	
		super.setSize(argW, argH);

		// calculate zoom amount
		zoomX = argW/origW;
		zoomY = argH/origH; 
		if (col == COLLISION_TYPE.RADIUS)
		{
			calcRadius();
		}
	}

	/// set general zoom
	void setZoom(float argZ)
	{
		setZoom(argZ, argZ);
	}

	/// set zoom X and Y amounts
	void setZoom(float argZX, float argZY)
	{
		zoomX = argZX;
		zoomY = argZY;

		super.setSize(zoomX*origW, zoomY*origH);

		if (col == COLLISION_TYPE.RADIUS)
		{
			calcRadius();
		}
	}

	/// give a certain sound to a frame
	void setSound(Sound snd, char[] animation, int argNum)
	{
		anim[animation].setSound(snd, argNum);
	}

	/// retrieve a certain sound from a frame
	Sound getSound(char[] animation, int argNum)
	{
		return anim[animation].getSound(argNum); 
	}

	/// set frame to display for time amount
	void setTime(int argT, char[] animation, int argNum)
	{
		anim[animation].setTime(argT, argNum);
	}

	/// gives the current time display of frame
	int getTime(char[] animation, int argNum)
	{
		return anim[animation].getTime(argNum);
	}

	/// remove an animation
	void removeAnim(char[] animation)
	{
		anim.remove(animation);
	}

	/// set current animation
	void setAnim(char[] animation)
	{
		currAnim = animation;
	}

	/// reset an animation
	void resetAnim(char[] animation)
	{
		anim[animation].reset();
	}

	/// get current animation
	char[] getAnim()
	{
		return currAnim; 
	}

	/// get current animation frame number
	int getAnimNum()
	{
		return anim[currAnim].getCurr;
	}

	/// move based on getAngle facing, -1 is backwards, 1 is forwards
	void moveDir(float argSpeed)
	{
		setX(getX + sin(-getAngle*DEGREE_TO_RADIAN)*-argSpeed);
		setY(getY + cos(-getAngle*DEGREE_TO_RADIAN)*-argSpeed);
	}

	/// move based on a velocity vector
	void moveVel(Point vel)
	{
		setX(getX + vel.getX);
		setY(getY + vel.getY);
	}

	/// get velocity based on angle and speed
	Point getVelocityDir(float argSpeed)
	{
		return Point(sin(-targetAngle*DEGREE_TO_RADIAN)*-argSpeed,
		             cos(-targetAngle*DEGREE_TO_RADIAN)*-argSpeed); 
	}

	/// set look gradient, the number of degrees to rotate as we face a point
	void setRotateBy(float argLG)
	{
		rotateBy = argLG;
	}

	/// return the value of the processPosition amount
	float getRotateBy() { return rotateBy; }

	/// set value of the rotation speed
	void setRotateSpeed(float argSpeed)
	{
		speedOfRot = argSpeed;
	}

	/// get value of rotation speed
	float getRotateSpeed() { return speedOfRot; }

	/// have sprite face towards the direction of X,Y
	void pointTo(float argX, float argY)
	{
		/// thank you, phr00t!
		targetAngle = 180 - (atan2(argX-getX, argY-getY) * RADIAN_TO_DEGREE);

		// if rotateBy is just zero, then rotate directly there
		// otherwise it will be slowly rotated in 'process()'
		if (rotateBy == 0)
			angle = targetAngle;
	}

	/// have sprite face towards another sprite
	void pointTo(Sprite s)
	{
		pointTo(s.getX, s.getY);
	}

	/// just processes pivot points used for collision detection
	void processCollisionPos() 
	{
		calcPivot();
	}

	/// just processes animations
	void processAnimation() 
	{
		anim[currAnim].process();
	}

	/// just process rotation and turning
	void processRotation() 
	{
		rotSpeed.process(speedOfRot);

		// inRange function doesn't work here for some unknown reason
		if (withinRange(angle, targetAngle, rotateBy))
		{
			angle = targetAngle;
		}
		// not in range
		else if (
			rotateBy != 0 
			&& 
			rotSpeed.on 
			&& 
			!(angle == 0 && targetAngle == 360) 
			&& 
			!(targetAngle == 0 && angle == 360)
		   )
		{			
			float a = 360-targetAngle+angle;
			float b = targetAngle-angle;

			if (a > 360)
				a -= 360;

			if (b < 0)  
				b += 360;

			if (a >= b)  
			{
				angle += rotateBy;
				wrapAngle(angle);			
			}
			else if (a < b)
			{
				angle -= rotateBy;
				wrapAngle(angle);			
			}
		}   
	}

	/// rotate sprite left 
	void rotateLeft()
	{
		setAngle(getAngle-rotateBy);
	}

	/// rotate sprite right
	void rotateRight()
	{
		setAngle(getAngle+rotateBy);
	}

	/// process pivots, calculates pivot point and animations
	void process()
	{
		// calculate pivot point
		processCollisionPos();
		
		// animations
		processAnimation(); 
		
		// rotations
		processRotation();

		// rotation points
		processRotationPoint();
	}

	/// add a rotation point that will be kept track of by sprite
	void addRotationPoint(float argL, float argA)
	{
		// add rotation point with given length and getAngle
		rotPoint.length = rotPoint.length+1;
		rotPoint[rotPoint.length - 1].init(argL, argA-90);
	}

	/// get rotation point
	Point getRotationPoint(int argIndex)
	in
	{
		// bounds check 
		assert(argIndex < rotPoint.length && argIndex >= 0);
	}
	body
	{
		// return point
		return Point(rotPoint[argIndex].getX, rotPoint[argIndex].getY);
	}

	/// process rotation points
	void processRotationPoint()
	{
		// calculate all rotation points
		for(int i = 0; i < rotPoint.length; i++)
		{
			Point point = Point(rotPoint[i].length, rotPoint[i].getAngle+getAngle);
			point.polarToRect();
			rotPoint[i].getX = point.getX + getX;
			rotPoint[i].getY = point.getY + getY;
		}
	}


	// get leftmost getX value of sprite
	float getLeftX() { return leftX; }

	// get topmost getY value of sprite
	float getTopY() { return topY; }

	// return gap value
	float getGap() { return gap; }

  private:
	// calculate pivot point
	void calcPivot()
	{
		if (pivX != 0 && pivY != 0)
		{
			// calculate rotations based on pivot points
			Point p1;
			Point p2;

			p1 = Point(pivX,getAngle);
			p2 = Point(pivY,getAngle);
		
			p1.polarToRect();
			p2.polarToRect(); 

			// points with calculated pivot point
			leftX = getX-getHalfWidth-p2.getY+p1.getX;
			topY = getY-getHalfHeight+p2.getX+p1.getY;
		}
		else
		{
			// these will put it in the right position
			leftX = getX-getHalfWidth;
			topY = getY-getHalfHeight;
		}
	}

	// calculate radius of sprite
	void calcRadius()
	{
		radius = (getWidth+getHeight)/4;
	}


  private:  
	// animation information
	COLLISION_TYPE col; 
	Animation[char[]] anim;
	char[] currAnim = "default";
	float angle = 0;
	float gap = 0;

	// zoom var
	double zoomX = 1, zoomY = 1;

	// special rotational values
	float pivX=0, pivY=0;
	float leftX=0, topY=0; 

	// special vars for radius
	float radius=0;

	// special vars for pixframe
	float origW, origH;

	// number of degrees to rotate each time as we point to an object
	float rotateBy = 0;
	// target getAngle we are trying to reach
	float targetAngle = 0;
	// speed of rotation, time based
	float speedOfRot = .1;
	// rotation timer
	Blinker rotSpeed;

	// rotation points that sprite can keep track of
	RotationPoint[] rotPoint;
}

// points that sprite will keep track of and rotate
struct RotationPoint
{
	void init(float argL, float argA) { length = argL; getAngle = argA; }
	
	float length;
	float getAngle;
	float getX;
	float getY;
}