/*******************************************************************************

Description: OpenGL Texture Loading made easy

Authors: Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
--------------------
import arc.gfx.texture; 

int main() {
   // use our texture loading function to load texture and grab ID
   // which is used by OpenGL to identify the said texture
   Texture tex = arc.gfx.texture.texture("texture.png/jpg/tga/bmp/gif/pcx/lbm/xpm/pnm");

   // bind texture to whatever polygons you want 
   glBindTexture(GL_TEXTURE_2D, tex.getID);
      
   // draw said polygon
   glBegin(GL_QUADS);
      glTexCoord2f(in_left,  in_top);        glVertex2f(x,            		y);
      glTexCoord2f(in_right, in_top);        glVertex2f(x + tex.getWidth, 	y);
      glTexCoord2f(in_right, in_bottom);     glVertex2f(x + tex.getWidth, 	y + tex.getHeight);
      glTexCoord2f(in_left,  in_bottom);     glVertex2f(x,					y + tex.getHeight);
   glEnd(); 

   return 0;
}
--------------------

*******************************************************************************/
module arc.gfx.texture;

import derelict.sdl.sdl,
		derelict.sdl.image,
		derelict.opengl.gl, 
		derelict.opengl.glu;

import  arc.gfx.gfxutil,
		arc.phy.point, 
		arc.io.log;

import std.file, std.string;

// initialize module variables on module load up 
private 
{
	// keeps track of the texture number we are on
	uint gTextureCount = 0; 

	// store a single copy of each texture in memory, prevent multiple textures 
	// form being loaded more than once 
	Texture[char[]] textureList; 
}

public 
{
	/// jump texture count 
	void incrementTextureCount(uint argInc) { gTextureCount += argInc; }
}

/// Texture Data we hand back to the user
struct Texture
{
  public:
	/// texture 'constructor', load texture given full file name
	static Texture opCall(char[] texFileName) 
	{
		// if texture hasn't been loaded before, load it 
		if (!(texFileName in textureList))
		{
			// load texture 
			Texture t;
			t.loadTexture(texFileName); 
			textureList[texFileName] = t; 
			textureList.rehash; 
			return textureList[texFileName];
		}

		// otherwise just give the texture that was already loaded 
		return textureList[texFileName]; 
	}

  // hide loadTexture because using it by itself is not 'resource safe' 
  // you may end up with the same resource loaded twice 
  private: 
	/// load texture 
	void loadTexture(char[] texFileName)
	{
		// If our resource doesn't exist, report it and stop
		if (!std.file.exists(texFileName))
		{
			arc.io.log.write("texture", "loadTexture", "error", "Texture " ~ texFileName ~ " does not exist!"); 
			return; 
		}

		// read SDL image 
		SDL_Surface* image = IMG_Load(toStringz(texFileName));

		if (image is null)
		{
			arc.io.log.write("texture", "loadTexture", "error", "failed to load image"); 
            return 0; 
        }

		// convert image to 32 bit RGBA image if needed 
		if ((image.format.BitsPerPixel != 32) && (image.format.BitsPerPixel != 8))
		{
			// create new surface same size as old one but with 32 bit RGBA ordering
			SDL_Surface* newsurf =	SDL_CreateRGBSurface (SDL_SWSURFACE, 
											image.w, image.h, 32, 0x000000ff,
											0x0000ff00, 0x00ff0000, 0xff000000);

			// copy old surface data onto the new surface
			SDL_BlitSurface (image, null, newsurf, null);

			// free old surface memory 
			SDL_FreeSurface(image);

			// point the old surface data to the new surface 
			image = newsurf; 
		}

		// increase our texture count
		gTextureCount++; 

		// Bind the texture to the texture arrays index and init the texture
		glBindTexture(GL_TEXTURE_2D, gTextureCount);
		glPixelStorei(GL_UNPACK_ALIGNMENT,1);
   
		// Build Mipmaps (builds different versions of the picture for distances - looks better)
		gluBuild2DMipmaps(	GL_TEXTURE_2D, 4, image.w, image.h, 
							GL_RGBA, GL_UNSIGNED_BYTE, image.pixels);
                     
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

		// free image data
		SDL_FreeSurface(image); 

		arc.io.log.write("texture", "loadTexture", "info", "textured ", gTextureCount, " ", texFileName);

		// set texture size and ID attributes 
		width = image.w;
        height = image.h;
        
		ID = gTextureCount; 
	}

  public: 
	/// get size of the texture 
	float getWidth() { return width; }

    /// get height of the texture 
	float getHeight() { return height; }
    
	/// texture ID
	int getID() { return ID; }

  private:
	int ID;
	float width, height; 
}

/// Texture based on filename alone and return texture 
Texture texture(char[] argFileName) 
{
	return Texture(argFileName); 
}



