/*******************************************************************************

Description: A simplified sprite class with polygon collision detection meant for 
				GUI programming 

Authors: Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

*******************************************************************************/
module arc.gui.guisprite.guisprite;

private import std.math; 

private import 	arc.gfx.entity,
				arc.gfx.prim,
				arc.gfx.texture,
				arc.gui.guisprite.frame,
				arc.gui.guisprite.animation,
				arc.gui.guisprite.polygon,
				arc.phy.staticcollision,
				arc.phy.phyutil,
				arc.phy.time,
				arc.phy.point,
				arc.io.xml.xml,
				arc.io.sound,
				arc.io.log;
				
private import derelict.opengl.gl,
               derelict.sdl.sdl;

private import 	std.string,
				std.stream,
				std.file,
				std.conv;

const int DEFAULT_TIME = -1;
const char[] DEFAULT_ANIM = "default";

enum AUTOPOLY 
{
	// Create an automatic polygon box for each frame 
	BOX=0,

	// Create an automatic polygon circle with calculated radius for each frame 
	CIRCLE,

	NONE 
}

//////////////////////////////////////

/// A graphical object with animation and sound
class GUISprite : Entity
{
  public:
	/// initialize sprite and let programmer do the rest
	this()
	{
		polyType = AUTOPOLY.NONE; 
		create();
	}

	/// initialize sprite with automatic polygon type 
	this(AUTOPOLY argPolyType)
	{
		polyType = argPolyType; 
		create();
	}

	/// create sprite based on this polygon 
	this(Polygon p)
	{
		polygon = p; 
	}

	// cleanup
	~this()  
	{
		destroy();
	}

	// what everything needs to do to initialize itself
	void create()
	{
		zoom = Point(1,1);
	}

	void destroy()
	{
		// delete all animations
		foreach(char[] k; anim.keys)
			anim.remove(k);
	}

	/// add a frame to the animation
	void addFrame(char[] textureName, char[] animation, int time, Sound snd)
	{
		Texture tex; 
		
		if (std.file.exists(textureName))
		{
			tex = Texture(textureName);
		}
		else
		{
			arc.io.log.write("guisprite", "addFrame", "info", textureName, " does not exist"); 
			arc.io.log.flush();
		}
		
		// allocate new animation as required
		if (!(animation in anim)) {
			anim[animation] = new Animation;
			anim.rehash;
		}
	
		// otherwise just add it and set curr animation to it
		currAnim = animation;

		// load frame and set size
		anim[animation].addFrame(tex, time, snd);

		// only set size if a size is not already given 
		// will only be called for the first frame added 
		if (getWidth == 0 || getHeight == 0)
		{
			// and set our sprite size to it
			origSize = Point(tex.getWidth, tex.getHeight); 

			setSize(tex.getWidth, tex.getHeight);

			// if AUTOPOLY circle or box is given
			// create polygon based off of it
			if (polyType == AUTOPOLY.BOX)
			{
				polygon.destroy(); 
				polygon.createBox(tex.getWidth, tex.getHeight); 
			}
			else if (polyType == AUTOPOLY.CIRCLE)
			{
				polygon.destroy(); 
				float radius = (tex.getWidth+tex.getHeight)/4;
				polygon.createCircle(radius, cast(int)radius*3); 
			}
		}
	}

	/// returns true if point getX, getY collides with sprite
	bool collide(float mouseX, float mouseY)
	{
		return polygon.collide(mouseX, mouseY); 
	}

	/// overloaded mouse over 
	bool mouseOver()
	{
		return polygon.collide(arc.io.input.mouseX, arc.io.input.mouseY); 
	}

	/// collide a sprite with any other sprite
	bool collide(GUISprite s)
	{
		return polygon.collide(s.polygon);
	}

	/// draw frame 
	void draw()
	{	
		if (currAnim in anim)
			anim[currAnim].draw(getPosition, getSize, getR, getG, getB, getA);
	}

	/// draw the boundry of given type, won't draw anything for 'per-pixel'
	void drawBounds(ushort r, ushort g, ushort b, ushort a)
	{
		polygon.draw(r, g, b, a); 
	}

	/// return the current frame in the sprite
	Frame getCurrFrame()
	{
		return anim[currAnim].getCurrFrame;
	}

	/// set animation loop/not loop
	void setAnimLoop(char[] animation, bool v)
	{
		if (animation in anim)
			anim[animation].setLoop(v);
	}

	/// get zoom X and Y values 
	Point getZoom() { return zoom; }

	/// set getWidth of sprite
	void setWidth(float argW)
	{
		setSize(argW, getHeight);
	}

	/// set getHeight of sprite
	void setHeight(float argH)
	{
		setSize(getWidth, argH);
	}

	/// set size of sprite
	void setSize(float argW, float argH)
	{	
		Point argSize = Point(argW, argH); 
	
		// 1) Calculate new zoom amount, calculated as ( new_size / old_size )
		zoom = argSize / origSize; 

		// 2) Set to the new zoom amount, which in turn sets the size of the sprite 
		setZoom(zoom); 
	}

	/// set general zoom
	void setZoom(Point argZ)
	in 
	{
		assert(argZ.getX >= 0 && argZ.getY >= 0);
	}
	body 
	{
		zoom = argZ; 

		Point p = origSize * argZ; 
		super.setSize(p.x, p.y);	
	}

	/// give a certain sound to a frame
	void setSound(Sound snd, char[] animation, int argNum)
	{
		anim[animation].setSound(snd, argNum);
	}

	/// retrieve a certain sound from a frame
	Sound getSound(char[] animation, int argNum)
	{
		return anim[animation].getSound(argNum); 
	}

	/// set frame to display for time amount
	void setTime(int argT, char[] animation, int argNum)
	{
		anim[animation].setTime(argT, argNum);
	}

	/// gives the current time display of frame
	int getTime(char[] animation, int argNum)
	{
		return anim[animation].getTime(argNum);
	}

	/// remove an animation
	void removeAnim(char[] animation)
	{
		anim.remove(animation);
	}

	/// set current animation
	void setAnim(char[] animation)
	{
		if (animation in anim)
			currAnim = animation;
	}

	/// reset an animation
	void resetAnim(char[] animation)
	{
		anim[animation].reset();
	}

	/// get current animation
	char[] getAnim()
	{
		return currAnim; 
	}

	/// get current animation frame number
	int getAnimNum()
	{
		return anim[currAnim].getCurr;
	}

	/// just processes animations
	void processAnimation() 
	{
		// process only if the animation exists 
		if (currAnim in anim)
		{
			anim[currAnim].process(getPosition, zoom);
		}
	}

	/// process zoom and animations
	void process()
	{
		// animations
		processAnimation(); 

		// process polygon 
		polygon.transformPoints(getPosition, zoom);
	}

  private:  
	// animation information
	Animation[char[]] anim;
	char[] currAnim = "default";

	// zoom amount
	Point zoom; 

	// holds the size of the first image given
	// used to proportion zooms 
	Point origSize;

	// polygon types 
	AUTOPOLY polyType = AUTOPOLY.NONE;  

	// polygon outline 
	Polygon polygon; 
}


