﻿// This module is dedicated to code reading the GUI xml data
module arc.gui.readxml; 

// XML parsing into Node structure code 
private import arc.io.xml.xml;

// stream file and text->int conversion utilites 
private import 	std.file,
				std.stream,
				std.stdio,
				std.string,
				std.conv; 

// Read Skin Data ////////////////////////////////////

// public interface, skin reader and skin class 
public
{
	// reads XML data from filename and returns a skin class 
	Skin readSkinXMLData(char[] fileName)
	{
		assert(std.file.exists(fileName)); 

		// initialize variables
		initializeValidParents();

		// Load XML into XML data structure
		XmlNode xml = new XmlNode();
		File file = new File(fileName, FileMode.In);
		xml = readDocument(file);
		file.close();

		// Read XML data structure into an easily accessable skin class
		Skin skin = new Skin(xml.getAttribute("font")); 

		// Recurse through the entire tree and save data into skin
		recurseReadSkinXML(skin, xml);

		return skin;
	}

	// Read layout data 
	LayoutData readLayoutXMLData(char[] fileName)
	{
		// Load XML into XML data structure
		assert(std.file.exists(fileName)); 
		XmlNode xml = new XmlNode();
		File file = new File(fileName, FileMode.In);
		xml = readDocument(file);
		file.close();

		LayoutData layout; 

		// Read XML data structure into an easily accessable skin class
		layout.skin = xml.getAttribute("skin"); 
		layout.name = xml.getAttribute("name");
		layout.x = toFloat(xml.getAttribute("x"));
		layout.y = toFloat(xml.getAttribute("y"));
		layout.width = toFloat(xml.getAttribute("width"));
		layout.height = toFloat(xml.getAttribute("height"));

		// Recurse through the entire tree and save data into skin
		recurseReadLayoutXML(layout, xml);

		return layout;
	}


	// store all skin information in this skin data class
	class Skin
	{
	  public:
		this(char[] argFont)
		{
			fontName = argFont; 
		}

		void addFrame(char[] animation, char[] widget, char[] image, char[] soundfx, int time)
		{
			// add 1 to frame 
			frames[widget][animation].length = frames[widget][animation].length+1;
			int curr = frames[widget][animation].length-1;

			frames[widget][animation][curr] = FrameData(image, soundfx, time);
		}

		// get frame data for a specific widget 
		FrameData[][char[]] getWidgetFrames(char[] widget)
		{
			return frames[widget];
		}
	
		void print()
		{
			foreach(char[] widget; frames.keys)
			{
				foreach(char[] anim; frames[widget].keys)
				{
					writefln("Widget ", widget, " with animation ", anim);
					
					foreach (FrameData frame; frames[widget][anim])
						frame.print("\t"); 
				}
			}
		}

		char[] getFont() { return fontName; }
	  
	  private: 
		char[] fontName; 
		FrameData[][char[]][char[]] frames; // 'frame["button"]["hover"][0] = FrameData("image.png", 30, "sound.wav")..
	}


	// Layout data //////////////////////////////////////////
	struct LayoutData
	{
		// name of the layout 
		char[] name;

		// x and y, width and height
		float x, y, width, height;

		// name of the skin 
		char[] skin; 

		// layout widget data arrays
		ImageData[] images; 
		ButtonData[] buttons; 
		LabelData[] labels;
		TextfieldData[] textfields; 

		// returns true if we already have this image name 
		bool imageExist(char[] imgName)
		{
			foreach(ImageData i; images)
				if (i.name == imgName)
					return true;

			return false; 
		}
		
		void addImage(char[] name, float x, float y, float width, float height, char[] color, char[] animation, char[] image, char[] soundfx, int speed)
		{
			images.length = images.length+1;
			int curr = images.length-1; 

			images[curr].name = name;
			images[curr].x = x;
			images[curr].y = y;
			images[curr].width = width;
			images[curr].height = height; 
			images[curr].color = color; 
			images[curr].addFrame(animation, image, soundfx, speed); 
		} 

		// add new frame to current image 
		void addImageFrame(char[] animation, char[] image, char[] soundfx, int speed, bool newImg=true)
		{
			int curr = images.length-1; 
			images[curr].addFrame(animation, image, soundfx, speed); 
		}		

		void addLabel(float x, float y, uint fontheight, uint padding, char[] name, char[] color, char[] fontcolor)
		{
			labels.length = labels.length+1;
			int curr = labels.length-1; 

			labels[curr].x = x;
			labels[curr].y = y;
			labels[curr].fontheight = fontheight;
			labels[curr].padding = padding;
			labels[curr].name = name;
			labels[curr].color = color;
			labels[curr].fontcolor = fontcolor; 
		}

		void setCurrLabelText(char[] argT)
		{
			labels[labels.length-1].text = argT; 
		}

		void addTextField(float x, float y, uint maxWidth, uint fontheight, uint padding, uint maxlines, char[] name, char[] color, char[] fontcolor)
		{
			textfields.length = textfields.length+1;
			int curr = textfields.length-1; 

			textfields[curr].x = x;
			textfields[curr].y = y;
			textfields[curr].maxwidth = maxWidth; 
			textfields[curr].fontheight = fontheight; 
			textfields[curr].padding = padding; 
			textfields[curr].maxlines = maxlines; 
			textfields[curr].name = name; 
			textfields[curr].color = color; 
			textfields[curr].fontcolor = fontcolor; 
		}

		void setCurrTextfieldText(char[] argT)
		{
			textfields[textfields.length-1].text = argT; 
		}

		void addButton(float x, float y, uint fontheight, uint padding, char[] name, char[] color, char[] fontcolor)
		{
			buttons.length = buttons.length+1;
			int curr = buttons.length-1; 

			buttons[curr].x = x;
			buttons[curr].y = y;
			buttons[curr].fontheight = fontheight;
			buttons[curr].padding = padding;
			buttons[curr].name = name;
			buttons[curr].color = color;
			buttons[curr].fontcolor = fontcolor; 
		}

		void setCurrButtonText(char[] argT)
		{
			buttons[buttons.length-1].text = argT; 
		}

		void print()
		{
			int curr = 1;
			foreach(ImageData img; images)
			{
				writefln("Image num ", curr++); 
				img.print(); 
			}
		}
	}
	

	struct ButtonData
	{
		float x, y;
		uint fontheight; 
		uint padding; 
		char[] name, text, color, fontcolor; 
	}

	struct TextfieldData 
	{
		float x, y, width; 
		uint fontheight; 
		uint padding; 
		uint maxwidth;
		uint maxlines; 
		char[] name, text, color, fontcolor;
	}

	struct LabelData 
	{
		float x, y;
		uint fontheight; 
		uint padding; 
		char[] name, text, color, fontcolor; 
	}

	// single image can hold a list of frames 
	struct ImageData
	{
		ImageFrameData[] frames;
		float x, y, width, height;
		char[] name; 
		char[] color; 

		void addFrame(char[] animation, char[] image, char[] soundfx, int speed)
		{
			frames.length = frames.length+1;
			int curr = frames.length-1; 

			frames[curr].animation = animation; 
			frames[curr].image = image; 
			frames[curr].soundfx = soundfx;
			frames[curr].speed = speed;
		}

		void print()
		{
			writefln("Image x is ", x, " and y is ", y, " with name ", name); 

			foreach(ImageFrameData image; frames)
				image.print(); 
		}
	}

	struct ImageFrameData
	{
		char[] animation, image, soundfx; 
		int speed; 

		void print()
		{
			writefln("Animation is ", animation, " with image ", image, " and soundfx ", soundfx, " and speed ", speed);
		}
	}
	
}

// privates user should not access 
private 
{
	// skin vars
	char[][] validWidgets; 
	char[][] validSkinAnims; 
	char[] currSkinWidget; 
	char[] currSkinAnim; 

	// layout vars
	float currLayoutImageX;
	float currLayoutImageY;
	float currLayoutImageWidth=0;
	float currLayoutImageHeight=0;
	char[] currLayoutImageName;
	char[] currLayoutImageColor; 
	
	bool newImg = false; 

	void initializeValidParents()
	{
		validWidgets.length = 4; 
		validWidgets[0] = "button"; 
		validWidgets[1] = "image"; 
		validWidgets[2] = "label";
		validWidgets[3] = "textfield";

		validSkinAnims.length = 5; 
		validSkinAnims[0] = "hover"; 
		validSkinAnims[1] = "click"; 
		validSkinAnims[2] = "unclick"; 
		validSkinAnims[3] = "default"; 
		validSkinAnims[4] = "focus";
		
	}


	// Recursively print all Nodes in the xml tree
	void recurseReadSkinXML(Skin skin, XmlNode node, char[] parentName="")
	{
		// For each xml node 
		foreach (XmlNode child; node.getChildren())
		{
			// loop through possibilities of valid parents 
			foreach(char[] possibleWidget; validWidgets)
			{
				// we find a match then we set current widget 
				if (child.getName() == "animation")
				{
					if (parentName == possibleWidget)// && child.getName() == "animation")
					{
						currSkinWidget = parentName; 

						// loop through valid possibile animation 
						foreach(char[] possibleAnim; validSkinAnims)
						{
							// we find a match then we set current animation 
							if (child.getAttribute("name") == possibleAnim)
							{
								currSkinAnim = possibleAnim; 
							}
						}
					}
				}
			}

			// add frame data when we encounter it
			if (parentName == "animation" && child.getName() == "frame")
			{
				skin.addFrame(currSkinAnim, currSkinWidget, 
				child.getAttribute("image"),
				child.getAttribute("soundfx"),
				std.conv.toInt(child.getAttribute("speed")));
			}

			// recurse until we have read entire structure 
			recurseReadSkinXML(skin, child, child.getName());
		}

	}


	// Recursively print all Nodes in the xml tree
	void recurseReadLayoutXML(inout LayoutData layout, XmlNode node, char[] parentName="")
	{
		// For each xml node 
		foreach (XmlNode child; node.getChildren())
		{
			char[] name = child.getName(); 
			
			if (name == "image")
			{
				currLayoutImageX = toFloat(child.getAttribute("x"));
				currLayoutImageY = toFloat(child.getAttribute("y"));
				currLayoutImageWidth = toFloat(child.getAttribute("width"));
				currLayoutImageHeight = toFloat(child.getAttribute("height"));
				currLayoutImageName = child.getAttribute("name");
				currLayoutImageColor = child.getAttribute("color"); 

				newImg = true; 
			}
			else if (name == "frame")
			{
				// add a new image if it is a new image 
				if (newImg)
				{
					layout.addImage(currLayoutImageName, 
									currLayoutImageX,
									currLayoutImageY, 
									currLayoutImageWidth, 
									currLayoutImageHeight,
									currLayoutImageColor, 
									child.getAttribute("animation"), 
									child.getAttribute("image"),
									child.getAttribute("soundfx"),
									toInt(child.getAttribute("speed")));
				}
				// otherwise just add a new frame 
				else
				{
					layout.addImageFrame(	child.getAttribute("animation"),
											child.getAttribute("image"),
											child.getAttribute("soundfx"),
											toInt(child.getAttribute("speed"))); 
				}

				// it is not a new image anymore until the next image is called 
				newImg = false; 
			}
			else if (name == "label")
			{
				layout.addLabel(toFloat(child.getAttribute("x")),
								toFloat(child.getAttribute("y")),
								toUint(child.getAttribute("fontheight")),
								toUint(child.getAttribute("padding")),
								child.getAttribute("name"),
								child.getAttribute("color"),
								child.getAttribute("fontcolor")); 
			}
			else if (name == "button")
			{
				layout.addButton(toFloat(child.getAttribute("x")),
								toFloat(child.getAttribute("y")),
								toUint(child.getAttribute("fontheight")),
								toUint(child.getAttribute("padding")),
								child.getAttribute("name"),
								child.getAttribute("color"),
								child.getAttribute("fontcolor")); 
			}
			else if (name == "textfield")
			{
				layout.addTextField(toFloat(child.getAttribute("x")), 
									toFloat(child.getAttribute("y")), 
									toUint(child.getAttribute("maxwidth")), 
									toUint(child.getAttribute("fontheight")), 
									toUint(child.getAttribute("padding")), 
									toUint(child.getAttribute("maxlines")), 
									child.getAttribute("name"), 
									child.getAttribute("color"), 
									child.getAttribute("fontcolor"));

			}

			// child has Cdata text
			if (child.isCdata())
			{
				
				if (parentName == "label")
				{
					// split by newlines
					char[][] lines = split(child.getCdata, r"\n");
					char[] newtext=""; 

					foreach(char[] line; lines)
					{
						newtext ~= line ~ '\n';
					}
					
					layout.setCurrLabelText(newtext); 
				}	
				else if (parentName == "button")
				{
					// split by newlines
					char[][] lines = split(child.getCdata, r"\n");
					char[] newtext=""; 

					foreach(char[] line; lines)
					{
						newtext ~= line ~ '\n';
					}
					
					layout.setCurrButtonText(newtext); 
				}
				else if (parentName == "textfield")
				{
					// split by newlines
					char[][] lines = split(child.getCdata, r"\n");
					char[] newtext=""; 

					foreach(char[] line; lines)
					{
						newtext ~= line ~ '\n';
					}
					
					layout.setCurrTextfieldText(newtext);
				}

			}

			// recurse until we have read entire structure 
			recurseReadLayoutXML(layout, child, child.getName());
		}

	}

	// stores data of an individual frame 
	struct FrameData
	{
	  public:
		static FrameData opCall(char[] argI, char[] argSnd, int argT)
		{
			FrameData f;
			f.image = argI; 
			f.soundfx = argSnd; 
			f.time = argT; 
			return f;
		}

		void print(char[] header)
		{
			writefln(header, "Image is ", image, " , sound is ", soundfx, ", and time is ", time); 
		}
	  
	   char[] image, soundfx;
	   int time;
	}


}
