/*******************************************************************************

   Captures input from the keyboard and the mouse and makes it 
   easily accessable. 
   
   Copyright: 2005 Clay Smith
   
   Date:       (2005): modified to include keyHit functionality
   
   Authors:        Clay Smith
   
   Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
-------------------------------------------------------------------------------
import arc.io.input;

int main() {
   
   // initializes input
   arc.io.input.open();

   // while the user hasn't closed the window
   while (!arc.io.input.keyDown(SDL_QUIT))
   {
      // get current input from user
      arc.io.input.process();

      // user lightly taps 't' key
      arc.io.input.keyHit('t'); // a-z, SDLK_a-SDLK_z
      
      // user holds down 't' key
      arc.io.input.keyDown('t');
   
      // user lightly clicks right mouse button
      arc.io.input.mouseHit(RIGHT);

      // user holds down left mouse button
      arc.io.input.mouseDown(LEFT) // RIGHT and MIDDLE work as well

      // returns true if user hits a character
      if (arc.io.input.charHit) {
         // returns the last character user hit
         char ch = arc.io.input.lastChar; 
      }

      // returns position of mouse
      arc.io.input.mouseX; // .mouseY  .mouseOldY  .mouseOldX 

      // returns true if mouse is in motion
      arc.io.input.mouseMotion; 
   
      // returns true if mouse is wheeling up
      arc.io.input.wheelUp; // wheelDown - returns true on mouse Wheelup and Wheeldown

      // return true if current modifier is down
      arc.io.input.modHit(LSHIFT); // RSHIFT, LCTRL, RCTRL, LALT, RALT, LMETA, RMETA, NUM, CAPS
	  arc.io.input.modDown(LSHIFT);
   }

   // there is no close() or shutdown() type function for input

   return 0;
}
-------------------------------------------------------------------------------
*******************************************************************************/
module arc.io.input;

// publicly import input key identifiers
public import derelict.sdl.keysym,
              derelict.sdl.events; 

// texture reload
import arc.gfx.texture;
// custom mouse cursor
import 	arc.gfx.frame.frame,
		arc.gfx.frame.boxframe;

// uses SDL to handle input
import derelict.sdl.sdl;

// Input uses window when user resizes the window
import arc.io.window;
import arc.io.log; 


//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      KEY DEFINITIONS
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

///   The infamous 'any' key used to represent any key
const int ANYKEY = ARC_UNKNOWN;
const int ARC_QUIT = 12; 

/// Arc keyboard keys
enum
{
    /* The keyboard syms have been cleverly chosen to map to ASCII */
    ARC_UNKNOWN        = 0,
    ARC_FIRST      = 0,
    ARC_BACKSPACE      = 8,
    ARC_TAB        = 9,
    ARC_CLEAR      = 12,
    ARC_RETURN     = 13,
    ARC_PAUSE      = 19,
    ARC_ESCAPE     = 27,
    ARC_SPACE      = 32,
    ARC_EXCLAIM        = 33,
    ARC_QUOTEDBL       = 34,
    ARC_HASH       = 35,
    ARC_DOLLAR     = 36,
    ARC_AMPERSAND      = 38,
    ARC_QUOTE      = 39,
    ARC_LEFTPAREN      = 40,
    ARC_RIGHTPAREN     = 41,
    ARC_ASTERISK       = 42,
    ARC_PLUS       = 43,
    ARC_COMMA      = 44,
    ARC_MINUS      = 45,
    ARC_PERIOD     = 46,
    ARC_SLASH      = 47,
    ARC_0          = 48,
    ARC_1          = 49,
    ARC_2          = 50,
    ARC_3          = 51,
    ARC_4          = 52,
    ARC_5          = 53,
    ARC_6          = 54,
    ARC_7          = 55,
    ARC_8          = 56,
    ARC_9          = 57,
    ARC_COLON      = 58,
    ARC_SEMICOLON      = 59,
    ARC_LESS       = 60,
    ARC_EQUALS     = 61,
    ARC_GREATER        = 62,
    ARC_QUESTION       = 63,
    ARC_AT         = 64,
    /*
       Skip uppercase letters
     */
    ARC_LEFTBRACKET    = 91,
    ARC_BACKSLASH      = 92,
    ARC_RIGHTBRACKET   = 93,
    ARC_CARET      = 94,
    ARC_UNDERSCORE     = 95,
    ARC_BACKQUOTE      = 96,
    ARC_a          = 97,
    ARC_b          = 98,
    ARC_c          = 99,
    ARC_d          = 100,
    ARC_e          = 101,
    ARC_f          = 102,
    ARC_g          = 103,
    ARC_h          = 104,
    ARC_i          = 105,
    ARC_j          = 106,
    ARC_k          = 107,
    ARC_l          = 108,
    ARC_m          = 109,
    ARC_n          = 110,
    ARC_o          = 111,
    ARC_p          = 112,
    ARC_q          = 113,
    ARC_r          = 114,
    ARC_s          = 115,
    ARC_t          = 116,
    ARC_u          = 117,
    ARC_v          = 118,
    ARC_w          = 119,
    ARC_x          = 120,
    ARC_y          = 121,
    ARC_z          = 122,
    ARC_DELETE     = 127,
    /* End of ASCII mapped keysyms */

    /* International keyboard syms */
    ARC_WORLD_0        = 160,      /* 0xA0 */
    ARC_WORLD_1        = 161,
    ARC_WORLD_2        = 162,
    ARC_WORLD_3        = 163,
    ARC_WORLD_4        = 164,
    ARC_WORLD_5        = 165,
    ARC_WORLD_6        = 166,
    ARC_WORLD_7        = 167,
    ARC_WORLD_8        = 168,
    ARC_WORLD_9        = 169,
    ARC_WORLD_10       = 170,
    ARC_WORLD_11       = 171,
    ARC_WORLD_12       = 172,
    ARC_WORLD_13       = 173,
    ARC_WORLD_14       = 174,
    ARC_WORLD_15       = 175,
    ARC_WORLD_16       = 176,
    ARC_WORLD_17       = 177,
    ARC_WORLD_18       = 178,
    ARC_WORLD_19       = 179,
    ARC_WORLD_20       = 180,
    ARC_WORLD_21       = 181,
    ARC_WORLD_22       = 182,
    ARC_WORLD_23       = 183,
    ARC_WORLD_24       = 184,
    ARC_WORLD_25       = 185,
    ARC_WORLD_26       = 186,
    ARC_WORLD_27       = 187,
    ARC_WORLD_28       = 188,
    ARC_WORLD_29       = 189,
    ARC_WORLD_30       = 190,
    ARC_WORLD_31       = 191,
    ARC_WORLD_32       = 192,
    ARC_WORLD_33       = 193,
    ARC_WORLD_34       = 194,
    ARC_WORLD_35       = 195,
    ARC_WORLD_36       = 196,
    ARC_WORLD_37       = 197,
    ARC_WORLD_38       = 198,
    ARC_WORLD_39       = 199,
    ARC_WORLD_40       = 200,
    ARC_WORLD_41       = 201,
    ARC_WORLD_42       = 202,
    ARC_WORLD_43       = 203,
    ARC_WORLD_44       = 204,
    ARC_WORLD_45       = 205,
    ARC_WORLD_46       = 206,
    ARC_WORLD_47       = 207,
    ARC_WORLD_48       = 208,
    ARC_WORLD_49       = 209,
    ARC_WORLD_50       = 210,
    ARC_WORLD_51       = 211,
    ARC_WORLD_52       = 212,
    ARC_WORLD_53       = 213,
    ARC_WORLD_54       = 214,
    ARC_WORLD_55       = 215,
    ARC_WORLD_56       = 216,
    ARC_WORLD_57       = 217,
    ARC_WORLD_58       = 218,
    ARC_WORLD_59       = 219,
    ARC_WORLD_60       = 220,
    ARC_WORLD_61       = 221,
    ARC_WORLD_62       = 222,
    ARC_WORLD_63       = 223,
    ARC_WORLD_64       = 224,
    ARC_WORLD_65       = 225,
    ARC_WORLD_66       = 226,
    ARC_WORLD_67       = 227,
    ARC_WORLD_68       = 228,
    ARC_WORLD_69       = 229,
    ARC_WORLD_70       = 230,
    ARC_WORLD_71       = 231,
    ARC_WORLD_72       = 232,
    ARC_WORLD_73       = 233,
    ARC_WORLD_74       = 234,
    ARC_WORLD_75       = 235,
    ARC_WORLD_76       = 236,
    ARC_WORLD_77       = 237,
    ARC_WORLD_78       = 238,
    ARC_WORLD_79       = 239,
    ARC_WORLD_80       = 240,
    ARC_WORLD_81       = 241,
    ARC_WORLD_82       = 242,
    ARC_WORLD_83       = 243,
    ARC_WORLD_84       = 244,
    ARC_WORLD_85       = 245,
    ARC_WORLD_86       = 246,
    ARC_WORLD_87       = 247,
    ARC_WORLD_88       = 248,
    ARC_WORLD_89       = 249,
    ARC_WORLD_90       = 250,
    ARC_WORLD_91       = 251,
    ARC_WORLD_92       = 252,
    ARC_WORLD_93       = 253,
    ARC_WORLD_94       = 254,
    ARC_WORLD_95       = 255,      /* 0xFF */

    /* Numeric keypad */
    ARC_KP0        = 256,
    ARC_KP1        = 257,
    ARC_KP2        = 258,
    ARC_KP3        = 259,
    ARC_KP4        = 260,
    ARC_KP5        = 261,
    ARC_KP6        = 262,
    ARC_KP7        = 263,
    ARC_KP8        = 264,
    ARC_KP9        = 265,
    ARC_KP_PERIOD      = 266,
    ARC_KP_DIVIDE      = 267,
    ARC_KP_MULTIPLY    = 268,
    ARC_KP_MINUS       = 269,
    ARC_KP_PLUS        = 270,
    ARC_KP_ENTER       = 271,
    ARC_KP_EQUALS      = 272,

    /* Arrows + Home/End pad */
    ARC_UP         = 273,
    ARC_DOWN       = 274,
    ARC_RIGHT      = 275,
    ARC_LEFT       = 276,
    ARC_INSERT     = 277,
    ARC_HOME       = 278,
    ARC_END        = 279,
    ARC_PAGEUP     = 280,
    ARC_PAGEDOWN       = 281,

    /* Function keys */
    ARC_F1         = 282,
    ARC_F2         = 283,
    ARC_F3         = 284,
    ARC_F4         = 285,
    ARC_F5         = 286,
    ARC_F6         = 287,
    ARC_F7         = 288,
    ARC_F8         = 289,
    ARC_F9         = 290,
    ARC_F10        = 291,
    ARC_F11        = 292,
    ARC_F12        = 293,
    ARC_F13        = 294,
    ARC_F14        = 295,
    ARC_F15        = 296,

    /* Key state modifier keys */
    ARC_NUMLOCK        = 300,
    ARC_CAPSLOCK       = 301,
    ARC_SCROLLOCK      = 302,
    ARC_RSHIFT     = 303,
    ARC_LSHIFT     = 304,
    ARC_RCTRL      = 305,
    ARC_LCTRL      = 306,
    ARC_RALT       = 307,
    ARC_LALT       = 308,
    ARC_RMETA      = 309,
    ARC_LMETA      = 310,
    ARC_LSUPER     = 311,      /* Left "Windows" key */
    ARC_RSUPER     = 312,      /* Right "Windows" key */
    ARC_MODE       = 313,      /* "Alt Gr" key */
    ARC_COMPOSE        = 314,      /* Multi-key compose key */

    /* Miscellaneous function keys */
    ARC_HELP       = 315,
    ARC_PRINT      = 316,
    ARC_SYSREQ     = 317,
    ARC_BREAK      = 318,
    ARC_MENU       = 319,
    ARC_POWER      = 320,      /* Power Macintosh power key */
    ARC_EURO       = 321,      /* Some european keyboards */
    ARC_UNDO       = 322,      /* Atari keyboard has Undo */

    /* Add any other keys here */

    ARC_LAST
}



/*******************************************************************************

   All of the buttons a mouse has (start at zero for array operations)
      0. Any button is clicked
      1. Left
      2. Middle
      3. Right

*******************************************************************************/
enum
{
   CLICKED = 0,
   LEFT,
   MIDDLE,
   RIGHT,
   MAXMOUSEBUTTON
}


//   Modifier enums, self descriptive, used for the modifier array.
enum
{
   LSHIFT = 0,
   RSHIFT,
   LCTRL,
   RCTRL,
   LALT,
   RALT,
   LMETA,
   RMETA,
   NUM,
   CAPS,
   MODE,
   MAXMODKEY
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      PUBLIC INTERFACE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
public {

/// clear's input, gets current modifiers, enables unicode and key repeat with default values
void open()
{
   // focus is true unless otherwise set
   focus = true;

   // clear all users keys
   keyboard.clear();
   mouse.button.clear();

   // get current modifiers
   modifiers.process();

   // Enable UNICODE translation for keyboard input
   SDL_EnableUNICODE(1);

   // Enable auto repeat for keyboard input
   SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}

/// SDL window has lost focus 
bit lostFocus() { return !focus; }

/// returns true when key is lightly tapped
bit keyHit(int keyNum)  { return keyboard.hit[keyNum];   }
/// returns true when user is pressing key down
bit keyDown(int keyNum) { return keyboard.down[keyNum];  }
/// returns true if user hits a character of the keyboard
bit charHit()           { return keyboard.charHit;       }
/// returns last characters hit by the user
char[] lastChars()         { return keyboard.lastChars;      }

/// returns true if user taps mouse (pistol effect)
bit mouseHit(int keyNum)   { return mouse.button.hit[keyNum];  }
/// returns true if user holds mouse down (chaingun effect)
bit mouseDown(int keyNum)  { return mouse.button.down[keyNum]; }

// NOTE: for the arc 2D game engine, mouse.x and mouse.y are only used 
//       to help determine the virtual coordinates.
//       for 3D games, you may want to use mouse.x and mouse.y
//       but for 2D games, there is never a need to, so mouseX and mouseX2D
//       will return the same values

/// mouseX position 
int mouseX()               { return cast(int)mouse.x2D;      }
/// mouseY position 
int mouseY()               { return cast(int)mouse.y2D;      }
/// old mouse X position 
int mouseOldX()            { return cast(int)mouse.oldX2D;   }
/// old mouse Y position 
int mouseOldY()            { return cast(int)mouse.oldY2D;   }
/// set custom mouse cursor
void setCursor(char[] argFile) { mouse.setCursor(argFile); }
/// release custom mouse cursor
void releaseCursor() { mouse.releaseCursor(); }
/// set mouse cursor visibility
void setCursorVisible(bool argV) { mouse.setCursorVisible(argV); }
/// get mouse cursor visibility
bool getCursorVisible() { return mouse.getCursorVisible(); }
/// toggle cursor visibility
bool toggleCursorVisible() { return mouse.toggleCursor(); }

/// returns true when mouse is moving
bit mouseMotion()          { return mouse.moving;              }
/// returns true when mouse is wheeling up
bit wheelUp()     { return mouse.button.wheelUp;   }
/// returns true when mouse is wheeling down
bit wheelDown()   { return mouse.button.wheelDown; }
/// returns true if modifier is 'hit'
bit modHit(int keyNum)  { return modifiers.modHit[keyNum]; }
/// returns true if modifier is 'down'
bit modDown(int keyNum) { return modifiers.modDown[keyNum];  }

/// force quit of application
void forceQuit() { keyboard.down[ARC_QUIT] = true; }

/// returns true if quitting 
bool isQuit() { return keyboard.down[ARC_QUIT]; }


/// Capture input from user
void process()
{
   /*******************************************************************************

      Declare our event, initialize 'hit' vars to get the once time 'hit' effect
      Also, modifiers don't need to be in the event loop

   *******************************************************************************/
   SDL_Event event;
   mouse.moving = false;
   mouse.button.clearHit();
   keyboard.clearHit();

   modifiers.process();

   // Keep on looping through our event until all events are accounted for
   while (SDL_PollEvent(&event))
   {
      // get the type of event
      switch (event.type)
      {
         case SDL_ACTIVEEVENT:
            if ( event.active.gain == 0 )
               focus = false;
            else
               focus = true;
         break; 

         case SDL_MOUSEBUTTONUP:
            // clear mouse keys when button up
            mouse.button.clearDown();

         break;

         case SDL_MOUSEBUTTONDOWN:
            // handle mouse keys when button is down
            mouse.button.handleKeyDown();

         break;

         case SDL_MOUSEMOTION:
            // handle mouse motion
            mouse.handleMotion(event);

         break;

         case SDL_KEYDOWN:
            // handle key down
            keyboard.handleKeyDown(event);

         break;

         case SDL_KEYUP:
            // handle key up, release keys
            keyboard.handleKeyUp(event);

         break;

         case SDL_QUIT:
            // set our SDL quit key to true
            keyboard.down[SDL_QUIT] = true;

         break;

         case SDL_VIDEORESIZE:
            // have the window be resized based upon user input
            arc.io.window.resize(event.resize.w, event.resize.h);
            arc.io.window.mode2D();

            // reload windows textures because Windows is special
			// and re-initialize OpenGL
			version(Windows)
			{
                //arc.gfx.texture.reload();
				arc.io.window.initGL();
			}

         break;

         default:
         break;

      } // switch(event)

   } // while(SDL_Poll(&event))

   // handle mouse wheel events
   mouse.button.handleWheel(event);

   // draw custom cursor if there is one
   mouse.draw();

} // process()

} // public

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      PRIVATE INTERNALS
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

private {

// whether or not input has focus
bit focus;
// self explainatory
KeyBoard keyboard;
Modifiers modifiers; // SHIFT, CAPS, NUMLOCK, etc.
Mouse mouse;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      MOUSE
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/*******************************************************************************

   Mouse structure to hold all the information in a mouse.

   a) The first set of coordinates, 'real world' coordinates, are used to tell
      where the mouse is no matter the size of the window.

   b) The '2D' coordinates, aka virtual coordinates, stay the same relative to the
      window, so the window can be resized and the relative coordinates stay the
      same, making them good for 2D gui's.

   c) The next bit tells us whether or not the mouse has been moving. It is useful
      so you can say, that if the mouse is moving, move a window with it.

   d) The mouse button holds all button info.

   e) WheelUp and WheelDown are whether or not mouse wheelup and wheeldown are
      being used.

   f) Custom mouse cursor graphic

*******************************************************************************/
struct Mouse
{
	// a
	int x = 0, y = 0;
	int oldX = 0, oldY = 0;

	// b
	float x2D = 0, y2D = 0;
	float oldX2D = 0, oldY2D = 0;

	// c
	bit moving = false;

	// d
	MouseButton button;

	// f
	//Frame graphic; 
	bool customVisible = false;

	void draw()
	{
		if (customVisible)
		{
			//assert(graphic !is null);
			//graphic.setPosition(x2D, y2D);
			//graphic.draw();
		}
	}
	
	// set mouse cursor to custom graphic
	void setCursor(char[] argFile)
	{
		// if one is already set, release it
		//if (graphic !is null)
		//	delete graphic;

		//graphic = new BoxFrame(argFile);
		
		customVisible = true;
		SDL_ShowCursor(SDL_DISABLE);
	}

	// set custom graphic visibility
	void setCursorVisible(bool argVis)
	{
		customVisible = argVis;
	}

	// toggle cursor
	bool toggleCursor()
	{
		customVisible = !customVisible;
		return customVisible;
	}

	// query custom graphic visibility
	bool getCursorVisible()
	{
		return customVisible;
	}

	// release custom graphic mouse cursor
	void releaseCursor()
	{
		customVisible = false;

//		delete graphic;
		SDL_ShowCursor(SDL_ENABLE);
	}
	
	/*******************************************************************************

		Properly handle mouse motion

		NOTE: 	Virtual coordinates make it so a smaller or bigger window
				keeps the same relative mouse coordinates, whether it be 800x600 or
				anything else they are used so you can click on gui objects when they
				are resized by OpenGL.

	*******************************************************************************/
	void handleMotion(inout SDL_Event event)
	{
		moving = true;

		oldX = x;
		oldY = y;

		x = event.button.x;
		y = event.button.y;

		oldX2D = x2D;
		oldY2D = y2D;

		x2D = x/(arc.io.window.getRealWidth/arc.io.window.getWidth);
		y2D = y/(arc.io.window.getRealHeight/arc.io.window.getHeight);
	}

} // struct Mouse

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      MOUSE BUTTON
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/*******************************************************************************

   Holds all the information of all the mouse buttons
      a) whether the button is hit, aka tapped
      b) whether the button is held down over long periods of time
      c) whether wheelUp or wheelDown is being used or not

*******************************************************************************/
struct MouseButton
{
   // a
   bit hit[MAXMOUSEBUTTON];

   // b
   bit down[MAXMOUSEBUTTON];

   // c
   bit wheelUp = false, wheelDown = false;


   /*******************************************************************************

      Clear all data in the hit bit array

   *******************************************************************************/
	void clearHit()
	{
		for (int i = 0; i < MAXMOUSEBUTTON; i++)
			hit[i] = false;
	}


   /*******************************************************************************

      Clear all data in the down bit array

   *******************************************************************************/
   void clearDown()
   {
      for (int i = 0; i < MAXMOUSEBUTTON; i++)
         down[i] = false;
   }


   /*******************************************************************************

      Clear all data in the button

   *******************************************************************************/
   void clear()
   {
      clearHit();
      clearDown();
   }


   /*******************************************************************************

      Properly deal with mouse wheel up and wheel down events. If wheelup is true,
      than that must mean wheel down is false, and vice versa. This is so wheelup
      or wheel down are not both on at the same time, and are both on for a limited
      time

   *******************************************************************************/
   void handleWheel(inout SDL_Event event)
   {
      wheelUp = false;
      wheelDown = false;

      if (event.button.button == SDL_BUTTON_WHEELUP)
         wheelUp = true;

      if (event.button.button == SDL_BUTTON_WHEELDOWN)
         wheelDown = true;
   }


   /*******************************************************************************

      Properly handle mouse when mouse is down

   *******************************************************************************/
   void handleKeyDown()
   {
      down[CLICKED] = true;

      if(SDL_GetMouseState(null, null)&SDL_BUTTON_LMASK)
      {
         hit[LEFT] = true;
         down[LEFT] = true;
      }

      if(SDL_GetMouseState(null, null)&SDL_BUTTON_MMASK)
      {
         hit[MIDDLE] = true;
         down[MIDDLE] = true;
      }

      if(SDL_GetMouseState(null, null)&SDL_BUTTON_RMASK)
      {
         hit[RIGHT] = true;
         down[RIGHT] = true;
      }
   }

} // struct MouseButton

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      MODIFIERS
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/*******************************************************************************

   Holds modifier keys such as Shift, CTRL, CAPS, NUM.
      a) bit array tells us whether modifier is down or not

*******************************************************************************/
struct Modifiers
{
   // a
   bit modDown[MAXMODKEY];
   bit modHit[MAXMODKEY];
   bit downLast[MAXMODKEY];

   /*******************************************************************************

      Query mod key and return true is mod key is pressed down

   *******************************************************************************/
   bit queryMod(int modKey, int modType)
   {
      if (modKey & modType)
         return true;
      
      return false;
   }


   /*******************************************************************************

      Process all modifier keys appropriately, setting them as pressed down
         or not.

   *******************************************************************************/
   void process()
   {
		// collect modifiers
		int modKey   =  SDL_GetModState();

		modDown[CAPS]    = queryMod(modKey, KMOD_CAPS);
		modDown[NUM]     = queryMod(modKey, KMOD_NUM);
		modDown[LSHIFT]  = queryMod(modKey, KMOD_LSHIFT);
		modDown[RSHIFT]  = queryMod(modKey, KMOD_RSHIFT);
		modDown[LCTRL] =   queryMod(modKey, KMOD_LCTRL);
		modDown[RCTRL] =   queryMod(modKey, KMOD_RCTRL);
		modDown[LALT]    = queryMod(modKey, KMOD_LALT);
		modDown[RALT]    = queryMod(modKey, KMOD_RALT);
		modDown[LMETA]   = queryMod(modKey, KMOD_LMETA);
		modDown[RMETA]   = queryMod(modKey, KMOD_RMETA);
		modDown[MODE]    = queryMod(modKey, KMOD_MODE);

		/// Give modifiers the 'keyHit' ability
		/// CAPS ///////////////
		if (modDown[CAPS] == true && downLast[CAPS] == false)
		{
			modHit[CAPS] = downLast[CAPS] = true;
		}
		else if (modDown[CAPS] == true && downLast[CAPS] == true)
		{
			modHit[CAPS] = false;
		}
		else
			downLast[CAPS] = false;

		/// NUM ///////////////
		if (modDown[NUM] == true && downLast[NUM] == false)
		{
			modHit[NUM] = downLast[NUM] = true;
		}
		else if (modDown[NUM] == true && downLast[NUM] == true)
		{
			modHit[NUM] = false;
		}
		else
			downLast[NUM] = false;

		/// LSHIFT ///////////////
		if (modDown[LSHIFT] == true && downLast[LSHIFT] == false)
		{
			modHit[LSHIFT] = downLast[LSHIFT] = true;
		}
		else if (modDown[LSHIFT] == true && downLast[LSHIFT] == true)
		{
			modHit[LSHIFT] = false;
		}
		else
			downLast[LSHIFT] = false;

		/// RSHIFT ///////////////
		if (modDown[RSHIFT] == true && downLast[RSHIFT] == false)
		{
			modHit[RSHIFT] = downLast[RSHIFT] = true;
		}
		else if (modDown[RSHIFT] == true && downLast[RSHIFT] == true)
		{
			modHit[RSHIFT] = false;
		}
		else
			downLast[RSHIFT] = false;

		/// LCTRL ///////////////
		if (modDown[LCTRL] == true && downLast[LCTRL] == false)
		{
			modHit[LCTRL] = downLast[LCTRL] = true;
		}
		else if (modDown[LCTRL] == true && downLast[LCTRL] == true)
		{
			modHit[LCTRL] = false;
		}
		else
			downLast[LCTRL] = false;

		/// RCTRL ///////////////
		if (modDown[RCTRL] == true && downLast[RCTRL] == false)
		{
			modHit[RCTRL] = downLast[RCTRL] = true;
		}
		else if (modDown[RCTRL] == true && downLast[RCTRL] == true)
		{
			modHit[RCTRL] = false;
		}
		else
			downLast[RCTRL] = false;

		/// LALT ///////////////
		if (modDown[LALT] == true && downLast[LALT] == false)
		{
			modHit[LALT] = downLast[LALT] = true;
		}
		else if (modDown[LALT] == true && downLast[LALT] == true)
		{
			modHit[LALT] = false;
		}
		else
			downLast[LALT] = false;

		/// RALT ///////////////
		if (modDown[RALT] == true && downLast[RALT] == false)
		{
			modHit[RALT] = downLast[RALT] = true;
		}
		else if (modDown[RALT] == true && downLast[RALT] == true)
		{
			modHit[RALT] = false;
		}
		else
			downLast[RALT] = false;

		/// LMETA ///////////////
		if (modDown[LMETA] == true && downLast[LMETA] == false)
		{
			modHit[LMETA] = downLast[LMETA] = true;
		}
		else if (modDown[LMETA] == true && downLast[LMETA] == true)
		{
			modHit[LMETA] = false;
		}
		else
			downLast[LMETA] = false;

		/// RMETA ///////////////
		if (modDown[RMETA] == true && downLast[RMETA] == false)
		{
			modHit[RMETA] = downLast[RMETA] = true;
		}
		else if (modDown[RMETA] == true && downLast[RMETA] == true)
		{
			modHit[RMETA] = false;
		}
		else
			downLast[RMETA] = false;

		/// MODE ///////////////
		if (modDown[MODE] == true && downLast[MODE] == false)
		{
			modHit[MODE] = downLast[MODE] = true;
		}
		else if (modDown[MODE] == true && downLast[MODE] == true)
		{
			modHit[MODE] = false;
		}
		else
			downLast[MODE] = false;

   } // process()

} // struct Modifiers

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                      KEYBOARD
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/*******************************************************************************

   Holds all data inside the keyboard
      a) when the user hits keyboard, or 'taps' it
      b) when the user holds key down for machine gun
      c) when the user types, capture last character, good for cheat codes
            or anything that the user ends up typing

*******************************************************************************/
struct KeyBoard
{
   // a
   bit hit[SDLK_LAST];

   //b
   bit down[SDLK_LAST];

   // c
   const char CHARNULL = cast(char)0;
	 // whether any character was hit between two process calls
   bit charHit = false;
	 // holds the characters hit between two process calls
   char[] lastChars;


   /*******************************************************************************

      Clear all data in the hit array

   *******************************************************************************/
   void clearHit()
   {
      charHit = false;
      lastChars.length = 0;
      for (int i = 0; i < SDLK_LAST; i++)
         hit[i] = false;
   }

   /*******************************************************************************

      clear all data in the down array

   *******************************************************************************/
   void clearDown()
   {
      for (int i = 0; i < SDLK_LAST; i++)
         down[i] = false;
   }

   /*******************************************************************************

      clear all data inside this structure

   *******************************************************************************/
   void clear()
   {
      clearHit();
      clearDown();
   }

   /*******************************************************************************

      Handle keyboard when keys are pressed

   *******************************************************************************/
   void handleKeyDown(inout SDL_Event event)
   {
		down[ANYKEY] = true; // anykey hit
		down[event.key.keysym.sym] = true;

		hit[ANYKEY] = true; // anykey hit
		hit[event.key.keysym.sym] = true;

		charHit = true;

		if ( event.key.keysym.unicode ) // if character needs to be translated
		{
			lastChars ~= cast(char)(event.key.keysym.unicode); // store the translated character
		}
		else
		{
		// note: this line causes weird bugs 
		//	lastChars ~= cast(char)(event.key.keysym.scancode);
		}
   }

   /*******************************************************************************

      Handle keyboard when keys are released

   *******************************************************************************/
   void handleKeyUp(inout SDL_Event event)
   {
      down[ANYKEY] = false; // all keys let go
      down[event.key.keysym.sym] = false; // not down anymore, whatever if was
   }
} // struct KeyBoard

 

} // private
