﻿/*******************************************************************************

Description: A Logger that can log to stdout, file, or xml file 

Authors: Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
--------------------
import arc.io.log; 

int main() 
{
	arc.io.log.open("stdout/file.txt/file.xml"); 

	arc.io.log.write("filename", "functionName", "errorLevel (info, warn, error), "message"); 

	arc.io.log.close(); 
	
	return 0;
}
--------------------

*******************************************************************************/

module arc.io.log; 

import 	arc.io.ioutil, arc.io.input,
		arc.io.xml.xml; 
		
import std.utf, std.string, std.stream; 

// this will be removed when full writef-->log transition is complete
public import std.stdio; 

// Log will have 4 levels: Info, Warning, Error, or Fault 
// Log will also need to know the filename and function name that the error came from
// Log will be stored as an array of LogMessage's, these lines will be sorted and outputed to 
// 	whichever location at program exit 
public
{
	/// open log and set where log will be sent to 
	void open(char[] argFile="")
	{
		if (argFile.length == 0)
		{
			outputType = "stdout";
			return;
		}

		outputType = argFile;
	}

	/// write log right now 
	void flush()
	{
		if (outputType == "stdout")
		{
			foreach(LogMessage m; lines)
			{
				m.stdout(); 
			}
		}
		else
		{
			char[][] ext = split(outputType, "."); 

			// write log to a textfile 
			if (ext[1] == "txt")
			{
				// open file for writing
				auto File s = new File(outputType, FileMode.Out); 

				// stream it to the file
				foreach(LogMessage m; lines)
					m.stream(s); 
					
			}
			// write log to an XML file 
			else if (ext[1] == "xml")
			{
				// Create XML structure with file to write to 
				auto XmlNode xml = new XmlNode("log"); 
				auto File s = new File(outputType, FileMode.Out); 

				// fill the XML structure with the log data 
				foreach(LogMessage m; lines)
					m.xml(xml); 

				// write XML data to stream 
				xml.write(s);
			}
		}

		// after we have 'flushed' the data, remove it
		lines.length = 0;
	}

	/// close log (causes a flush)
	void close()
	{
		flush(); 
	}

	/// Write a message to the log 
	void write(char[] filename, char[] functionName, char[] errorLevel, ...)
	in 
	{
		// force program to quit if 
		if (errorLevel != "info" && 
			errorLevel != "warn" && 
			errorLevel != "error")
		{
			// print to log the reason to quit
			lines.length = lines.length+1; 
			lines[lines.length-1] = LogMessage("log", "write", "error", "Error Level " ~ errorLevel ~ "is invalid"); 

			// force quit 
			arc.io.input.forceQuit(); 
		}
	}
	body 
	{
		// write the line into our logging system 
		lines.length = lines.length+1; 
		lines[lines.length-1] = LogMessage(filename, functionName, errorLevel, toUTF8(formatString(_arguments, _argptr))); 

		// force program to quit on errors 
		if (errorLevel == "error")
			arc.io.input.forceQuit(); 
	}
}

private 
{
	struct LogMessage
	{
		static LogMessage opCall(char[] file, char[] func, char[] err, char[] msg)
		{
			LogMessage m;
			m.filename = file;
			m.functionName = func; 
			m.errorLevel = err; 
			m.message = msg; 
			return m;
		}

		// write log message to stdout 
		void stdout()
		{
			writefln("(",errorLevel, ") | in file ", filename, ".d, function ", functionName, "()");
			writefln("Message:\n", message, "\n"); 
		}

		// write log message to filestream 
		void stream(Stream s)
		{
			char[] streamedText = "(" ~ errorLevel ~ ") | in file " ~ filename ~ ".d, function " ~ functionName ~ "()\n";
			streamedText ~= "Message:\n" ~ message ~ "\n"; 

			s.writeLine(streamedText); 
		}

		// write log to XML file 
		void xml(XmlNode xml)
		{
			// Create new XML child node with given log data 
			XmlNode child = new XmlNode("message"); 
			child.setAttribute("errorLevel", errorLevel);
			child.setAttribute("fileName", filename ~ ".d");
			child.setAttribute("functionName", functionName ~ "()");
			child.addCdata(message);

			// add it to the XML node 
			xml.addChild(child); 
		}
			
		char[] filename; 
		char[] functionName;
		char[] errorLevel; 
		char[] message;
	}

	LogMessage[] lines; 
	char[] outputType;
}
