/*******************************************************************************

Description:  The window module abstracts SDL's window initialization, 
               Derelict's dynamic loading, and OpenGL's window initialization
               

Authors: Clay Smith, clayasaurus@gmail.com

Copyright: 2005 Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>



Examples:
--------------------
import arc.io.window; 

int main() {
   // initialize SDL/OpenGL window 
   // read window parameters from a lua configuration file
   arc.io.window.open("Title", width, height, bFullScreen); 

   // take a screenshot of current window and save it as a bitmap
   arc.io.window.screenshot("screen1");
   
   // toggle between full screen and window mode (linux only)
   arc.io.window.toggleFullScreen();

   // resize window with given width and height (currently texture information is lost on Windows)
   arc.io.window.resize(width, height)

   // closes the window
   arc.io.window.close(); 

   return 0;
}
--------------------
      

*******************************************************************************/

module arc.io.window;

/*******************************************************************************

   Import all we need from the std library

*******************************************************************************/
private import std.string, // .cmp function
               std.c.stdlib, // exit() function
			   std.c.string; // memcpy()

import arc.io.log; 

/*******************************************************************************

   Import all we need from derelict publicly so we don't have to do it so much

*******************************************************************************/
import 	derelict.opengl.gl,
		derelict.opengl.glu,
		derelict.openal.al, 
		derelict.ogg.ogg,
		derelict.ogg.vorbis, 
		derelict.sdl.sdl,
		derelict.sdl.image,
		derelict.freetype.ft;

import derelict.util.exception;

// skips trying to load glXGetProcAddress if it could not find it
bool myMissingProcCallback(char[] libName, char[] procName)
{
   
   if(procName.cmp("glXGetProcAddress") == 0)    
      return true;                          

   return false;                           
}

/*******************************************************************************

   Private vars we hide inside the module

*******************************************************************************/
private
{
   const float wWidth2D = 800, wHeight2D = 600;

   SDL_Surface *screen = null;
   double fov = 45.0, zNear = .5, zFar = 2000.0;
   float wWidth = 800, wHeight = 600;
   int bpp = 0;
   bit wFullscreen = false;
   char[] wTitle;
}

public
{
	/// get the window screen
	SDL_Surface* getScreen() { return screen; }
	
   /*******************************************************************************

      a - Initialize variables   <BR>
      b - Load's Derelict Libraries needed by Arc at runtime; SDL image + mixer, OpenGL + GLU, Lua<BR>
      c - Create the SDL window based on configuration file and resize OpenGL to current window   <BR>
      d - Enter OpenGL's 2D drawing mode<BR>       

   *******************************************************************************/
   void open(char[] argTitle, int argWidth, int argHeight, bit argFullscreen)
   {
		// a
		wTitle = argTitle;
		wWidth = argWidth;
		wHeight = argHeight;
		wFullscreen = argFullscreen;
	   
		debug writefln("window: open(", wTitle, ", ", wWidth, ", ", wHeight, ", ", wFullscreen, ")");

		// b
		loadDerelict(); 

		// c
		initializeSDL();
		glResize(wWidth, wHeight);

		// d
		mode2D();

		debug printVendor();
   }

   /*******************************************************************************

      Close SDL window and delete the screen

   *******************************************************************************/
	void close()
	{
		SDL_Quit();
      
		DerelictGL.unload();
		DerelictGLU.unload();
		DerelictSDL.unload();
		DerelictAL.unload(); 
		DerelictALU.unload();
		DerelictOgg.unload(); 
		DerelictVorbis.unload(); 
		DerelictVorbisFile.unload();
		DerelictFT.unload();
   }

   /// initialize OpenGL parameters required to draw 
   void initGL()
   {
      glEnable(GL_TEXTURE_2D); 
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   }


   /// Returns windows virtual width (never changes once set)
   float getWidth() { return wWidth2D; }

   /// Returns windows virtual height (never changes once set)
   float getHeight() { return wHeight2D; }

   /// returns windows actual width
   float getRealWidth() { return wWidth; }

   /// returns windows actual height
   float getRealHeight() { return wHeight; }

   /// Returns true if window is fullscreen
   bit isFullScreen() { return wFullscreen; }

   /*******************************************************************************

      Resize window to desired width and height

   *******************************************************************************/
   void resize(float argWidth, float argHeight)
   {
      // a
      if (wFullscreen)
         screen = SDL_SetVideoMode(cast(int)argWidth, cast(int)argHeight, bpp, SDL_OPENGL|SDL_HWPALETTE|SDL_FULLSCREEN|SDL_RESIZABLE);
      else
         screen = SDL_SetVideoMode(cast(int)argWidth, cast(int)argHeight, bpp, SDL_OPENGL|SDL_HWPALETTE|SDL_RESIZABLE);

      // b
      glResize(argWidth, argHeight);
   }

   /*******************************************************************************
   
      Captures a screenshot and saves in BMP format to current directory

   *******************************************************************************/
   void screenshot(char[] argName)
   {
      // a
      SDL_Surface *image = null;
      SDL_Surface *temp = null;

      // b
      image = SDL_CreateRGBSurface(SDL_SWSURFACE, cast(int)wWidth, cast(int)wHeight, 24, 0x0000FF, 0x00FF00, 0xFF0000, 0);

      if(!(image is null))
      {
         // c
         temp  = SDL_CreateRGBSurface(SDL_SWSURFACE, cast(int)wWidth, cast(int)wHeight, 24, 0x0000FF, 0x00FF00, 0xFF0000, 0);

         if(!(temp is null))
         {
            // d
            glReadPixels(0, 0, cast(int)wWidth, cast(int)wHeight, GL_RGB, GL_UNSIGNED_BYTE, image.pixels);

            // e
            for (int idx = 0; idx < wHeight; idx++)
            {
               char* dest = cast(char *)(temp.pixels+3*cast(int)wWidth*idx);
               memcpy(dest, cast(char *)(image.pixels+3*cast(int)wWidth*(cast(int)wHeight-1-idx)), 3*cast(int)wWidth);
               endianswap(dest, 3, cast(int)wWidth);
            }

            // f
			argName ~= ".bmp";
            SDL_SaveBMP(temp, argName.ptr);
            SDL_FreeSurface(temp);
            debug writefln("window: screenshot " ~ argName ~ ".bmp taken");
         }

         SDL_FreeSurface(image);
      }

   } // screenshot()

   /*******************************************************************************

         a - Clear GL Projection Matrix <BR>
         b - Set drawing mode to 2D<BR>
         c - Clear GL Model View Matrix<BR>
         d - Disable depth and enable 2D opengl textures<BR>

   *******************************************************************************/
   void mode2D()
   {
      // a
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();

      // b
      glOrtho(0.0f, wWidth2D, wHeight2D, 0.0,-1.0f,1.0f);

      // c
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      // d
      glDisable(GL_DEPTH_TEST);     
      glEnable(GL_TEXTURE_2D);

	  initGL();
   }

   /*******************************************************************************

      Toggle between fullscreen and windowed mode; linux only

   *******************************************************************************/
   void toggleFullScreen()
   {
      if(SDL_WM_ToggleFullScreen(screen) == 0)
      {
         debug writefln("Window: Failed to toggle fullscreen");
         return;
      }

      wFullscreen = !wFullscreen;
   }



	/// clear the screen
	void clear()
	{
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	/// swap the screen buffers
	void swap()
	{
		SDL_GL_SwapBuffers(); 
	}

} // public

private
{
   /*******************************************************************************

      Performs an endian swap, used in screenshot

   *******************************************************************************/
   void endianswap(inout char *memory, int stride, int length)
   {
      if(*(cast(char *)&stride)) return;

      for (int w = 0; w < length; w++)
      {
         for (int i = 0; i < (stride/2); i++)
         {
            ubyte *p = cast(ubyte *)memory+w*stride;
            ubyte t = p[i];
            p[i] = p[stride-i-1];
            p[stride-i-1] = t;
         }
      }
   }

   /*******************************************************************************

      Initialize SDL
         a) check to see if SDL init is null
         b) initialize video
         c) set pixel depth and format
         d) set fullscreen flags properly
         e) set screen video mode (creates window)
         d) set window caption

   *******************************************************************************/
   void initializeSDL()
   {
      // for some mysterious reason this info is lost if we don't do this
      char* windowTitle = toStringz(wTitle);

      // a
      if(SDL_Init is null)
         throw new Exception("Window: SDL_Init is null");
 
      // b
      if(SDL_Init(SDL_INIT_VIDEO) < 0)
         throw new Exception("Window: Failed to initialize SDL Video");

      // c
      setupPixelDepth();
      setupPixelFormat();

      // d
      if (wFullscreen)
         screen = SDL_SetVideoMode(cast(int)wWidth, cast(int)wHeight, bpp,  SDL_OPENGL|SDL_HWPALETTE|SDL_FULLSCREEN);
      else
         screen = SDL_SetVideoMode(cast(int)wWidth, cast(int)wHeight, bpp,  SDL_OPENGL|SDL_HWPALETTE|SDL_RESIZABLE);

      // e
      if (screen is null)
         throw new Exception("Window: screen is null after SDL_SetVideoMode called");

      // d
      SDL_WM_SetCaption(windowTitle, null);
   }

   /*******************************************************************************

      Query's SDL for the current video info (desktop), extracts bpp and sets
         bpp to what it found it was, this way the color depth is the color depth
         the user uses from the desktop, and it should not crash

   *******************************************************************************/
   void setupPixelDepth()
   {
      SDL_VideoInfo *info = SDL_GetVideoInfo();

      if(info is null)
         throw new Exception("Window: SDL_GetVideoInfo() is null");

      bpp = info.vfmt.BitsPerPixel;
   }

   /*******************************************************************************

      Prints a slew of graphics driver and extension debug info

   *******************************************************************************/
   debug
   {
      void printVendor()
      {
		debug writefln("Render Type: OpenGL (Hardware)");
		printf( "Vendor     : %s\n", glGetString( GL_VENDOR ) );
		printf( "Renderer   : %s\n", glGetString( GL_RENDERER ) );
		printf( "Version    : %s\n", glGetString( GL_VERSION ) );
		printf( "Extensions : %s\n", glGetString( GL_EXTENSIONS ) );
      }
   }

   /*******************************************************************************

      Derelict is a library that allows me to load all my library functions at
         run time, and displaying a proper message if any single library fails 
         to load, a message like "Can't find library %s, go get it at %s"

      a) load OpenGL
      b) load GLU
      c) load SDL
      d) load SDL image
      e) load SDL mixer
      f) load lua
	  g) load freetype
      h) if any of these libs fail to load, display appropriate message to the user

   *******************************************************************************/
	void loadDerelict()
	{
		try {

			DerelictGL.load();
			DerelictGLU.load();
			DerelictSDL.load();
			DerelictSDLImage.load();
			DerelictAL.load(); 
			DerelictALU.load();
			DerelictOgg.load(); 
			DerelictVorbis.load(); 
			DerelictVorbisFile.load();
			DerelictFT.load();

		} // try

		// h
		catch (Exception e) {
			e.print();
			exit(0);
		} // catch

	} // loadDerelict

   /*******************************************************************************

      Sets up the pixel format the way we like it

   *******************************************************************************/
   void setupPixelFormat()
   {
      SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
      SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
      SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
   }


   /*******************************************************************************

      Resizes an openGL scene with vars we pass it
         a) set width and height to what we pass our function
         b) resize openGL scene

   *******************************************************************************/
   void glResize(float argWidth, float argHeight)
   {
      // a
      wWidth = argWidth;
      wHeight = argHeight;

      // b
      glResize();
   }

   /*******************************************************************************

      Resizes an openGL scene with vars we pass it
         a) prevent divide by zero error
         b) resize the viewports and load the identity matrix's

   *******************************************************************************/
   void glResize()
   {
      // a
      if (wHeight == 0)
         wHeight = 1;

      // b
      glViewport(0,0, cast(int)wWidth, cast(int)wHeight);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      gluPerspective(fov, wWidth / wHeight, zNear ,zFar);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
   }


} // private




