/*******************************************************************************

   A 2D Point structure
     
   Date: (2005): initial version
   
   Authors: Olivier Renault from gamedev.net, 
			- ported to D and additional functionality added by Clay Smith (clayasaurus)
			- point polygon collision done by <a href="http://www.ecse.rpi.edu/Homepages/wrf/Research/Short_Notes/pnpoly.html#The%20C%20Code">W. Randolph Franklin</a>
   
   Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
-------------------------------------------------------------------------------
import arc.phy.point;

int main() {
   
   Point vec = Point(x, y);

   return 0;
}
-------------------------------------------------------------------------------
*******************************************************************************/
module arc.phy.point; 

private import std.stdio, std.math, std.stream; 
private import derelict.opengl.gl;
private import arc.phy.phyutil, arc.io.serialization.serializer;

/// return sign of a float, -1 or 1
float sign(float x)
{
	return (x < 0.0f)? -1.0f : 1.0f;
}

/// swap float values
void swapf(inout float a, inout float b)
{
	float c = a;
	a = b;
	b = c;
}

/// clamp float value to min and max
float clampf(float x, float min, float max)
{
	return (x < min)? min : (x > max)? max : x;
}

/// wrap float value around min and max
float wrapf(float x, float min, float max)
{
	return (x < min)? (x - min) + max : (x > max)? (x - max) + min : x;
}

/// two pi
const float TWOPI = PI*2;  

const float RADTODEG = 180/PI; 
const float DEGTORAD = PI/180; 

/// convert radians to degrees
float radianToDegree(float rad)
{
	return rad * RADTODEG;
}

/// convert degrees to radians
float degreeToRadian(float deg)
{
	return deg * DEGTORAD;
}


/// A vector structure
struct Point
{
	float x=0, y=0; 

    /// Point 'constructor' 
	static Point opCall(float Ix,float Iy)
	{
		Point v;
		v.x = Ix;
		v.y = Iy;
		return v;
	}

	void print(char[] str="")
	{
		writefln(str, "X - ", x, ", Y - ", y);
	}

	float max() 
	{
		if (x > y)
			return x;
		return y;
	}

	float min() 
	{
		if (x < y)
			return x;
		return y;
	}

	Point opDivAssign( float Scalar)	{ x /= Scalar; y /= Scalar;		return *this; }
	Point opDiv (float  s)				{	return Point(x/s, y/s); }
	Point opDiv (Point s) { return Point(x/s.x, y/s.y); } 

	Point opMulAssign( float Scalar)	{ x *= Scalar; y *= Scalar;		return *this; }
	Point opMulAssign(Point v) { return Point(x*= v.x, y*= v.y); } 
	Point opMul (float  s)				{	return Point(x*s, y*s); }
	Point opMul (Point v) { return Point(x*v.x, y*v.y); }
	Point opMul (float k,  Point V) {	return Point(V.x*k, V.y*k); }
	
	 Point opAddAssign( Point Other) { x += Other.x;	y += Other.y;	return *this; }
	 Point opAdd ( Point V)		{	return Point(x+V.x, y+V.y); }
	 Point opAdd ( float V) { return Point(x+V, y+V); }
		
	 Point opSubAssign( Point Other) { x -= Other.x;	y -= Other.y;	return *this;	}
	 Point opSub ( Point V)		{	return Point(x-V.x, y-V.y); }
	 Point opSub ( float V) { return Point(x-V, y-V); }


	/// cross product
	 float crossProduct ( Point V)		{	return (x * V.y) - (y * V.x); } // cross product

    /// dot product
	float dotProduct ( Point V)		{	return (x*V.x) + (y*V.y); } // dot product

	/// perp product
	float perpProduct(Point V) { return (x*V.y) - (y*V.x); } 

	/// negative number 
	Point opNeg()  { return Point(-x, -y); }

	/// returns length of vector
	float length()  { return cast(float) sqrt(x*x + y*y); }

	/// normalises the Point
	float normalise() 
	{	
		float fLength = length();	
		
		if (fLength == 0.0f) 
			return 0.0f; 
		
		(*this) *= (1.0f / fLength); 
	
		return fLength;	
	}

	/// direction of the vector
	Point direction() 
	{
		Point Temp = *this;

		Temp.normalise();

		return Temp;
	}

	/// angle of the vector
	float angle( Point xE)
	{
		float dot = (*this).dotProduct(xE);
		float cross = (*this).crossProduct(xE);
		
		// angle between segments
		return cast(float) atan2(cross, dot);
	}

	/// simply rotate 
	Point getRotation(float angle)
	{
		angle *= DEGTORAD;  
		return Point	(x*cos(angle)-y*sin(angle),
						 x*sin(angle)+y*cos(angle));
	}

	/// rotate around pivot point 
	Point getRotation(Point center, float angle)
	{
		Point v = Point(x,y) + center;
		v.rotate(angle); 
		return v; 
	}	

	/// rotate the vector
	Point rotate(float angle)
	{
		angle *= DEGTORAD; 
		float tx = x;
		
		x =  x * cos(angle) - y * sin(angle);
		y = tx * sin(angle) + y * cos(angle);

		return *this;
	}

	/// rotate the vector around a center point
	Point rotate( Point xCentre, float fAngle)
	{
		Point D = *this - xCentre;
		D.rotate(fAngle);

		*this = xCentre + D;

		return *this;
	}

	/// clamp a vector to min and max values
	void clamp( Point min, Point max)
	{
		x = (x < min.x)? min.x : (x > max.x)? max.x : x;
		y = (y < min.y)? min.y : (y > max.y)? max.y : y;
	}

	/// random vector size between given ranges
	void randomise( Point xMin,  Point xMax)
	{
		x = random_range(cast(int)xMin.x, cast(int)xMax.x);
		y = random_range(cast(int)xMin.y, cast(int)xMax.y);
	}

	/// render the vector as a single point
	void render() 
	{
		glColor3f(1.0f, 1.0f, 0.1f);
		glPointSize(3.0f);
		glBegin(GL_POINTS);
		glVertex2f(x, y);
		glEnd();
	}


	final float getX() { return x; }
	final float getY() { return y; }

	/// same as x
	final float getWidth() { return x; }

	/// same as y
	final float getHeight() { return y; }

	final void setSize(float argW, float argH)
	{
		x = argW; y = argH;
	}

	/// set width
	final void setWidth(float argW) { x = argW; }

	/// set height
	final void setHeight(float argH) { y = argH; }

	/// set x
	final void setX(float argW) { x = argW; }

	/// set y
	final void setY(float argH) { y = argH; }

	/// add amount to X
	final void addX(float argV) { x += argV; }

	/// add amount to Y
	final void addY(float argV) { y += argV; }

	/// swap values with another point
	void swap(inout Point p)
	{
		Point tmp = Point(x, y);
		x = p.x;
		y = p.y;
		p = tmp;
	}

	/// returns true if point hasn't been initialized
	bool isNaN() 
	{ 
		if (isnan(x) || isnan(y))
			return true; 
			
		return false; 
	}

	/// assuming point is given polar coordinates, this will convert to rectangular coordinates
	void polarToRect() 
	{
		y = y * DEGREE_TO_RADIAN; // convert degrees to radian
   
		float x_save = x;
   
		x = x_save * cos(y); // i know, polar->y is used too much, but i'd like to eliminate the need
		y = x_save * sin(y); // for too many variables
	}

	/// return true if this point is above the other
	bool above(Point another)
	{
		if (y < another.y)
			return true;

		return false; 
	}

	/// calculate the distance between this point and the next
	real distance(Point another)
	{
		return sqrt(pow(another.x-x, 2.0) + pow(another.y-y, 2.0));
	}

	/// point is serializable
	void describe(T)(T s)
	{
		assert(s !is null);
		s.describe(x); 
		s.describe(y); 
	}

	/// convert 'local' to 'world' coordinates 
	Point localToWorldCoords(Point xWorldPosition, float fWorldOrientation)
	{
		rotate(fWorldOrientation);
		*this += xWorldPosition;
		return *this;
	}
}

