﻿/// static collision detection methods 
module arc.phy.staticcollision; 

// Static Box, Radius, and Polygon Collision detection algorithms 
import 	arc.phy.phyutil,
		arc.phy.point;
		
import std.math, std.stdio; 
        
// BOXES //////////////////////////////////////////////////////////////


/// whether 2 boxes collide with each other
bool boxBoxCollision(	float x1, float y1, float w1, float h1,
						float x2, float y2, float w2, float h2)
{
	// box's within x range
	if (( x1 > x2 ||  x1 + w1 > x2) && ( x1 < x2 + w2)) 
		// box's within y range
		if (( y1 > y2 ||  y1 + h1 > y2) && ( y1 < y2 + h2))    
			return true;

	return false;
}


/// whether a box and circle collide with each other
bool boxCircleCollision(float x, float y, float width, float height,
						float cx, float cy, float radius)
{
		// first we do a simple test to see if center of circle is inside of box
		if (boxXYCollision(cx, cy, x, y, width, height))
			return true;

		// this next algoritm will determine the distance from given edge 
		// of rectangle to the circle.
		// draw line from radframe (x,y) to gx, gy for a better 
		// understanding of what this algorithm does

		// destination x and y values this algorithms calculates
		float gx = 0, gy = 0; 

		// box center point
		float x2 = x + width/2;
		float y2 = y + height/2;

		// box bottom right corner
		float nw2 = x + width;
		float nh2 = y + height;

		// box frame is above radius frame
		if (y2 <= cy)
		{
			// middle
			if (cx >= x && cx <= nw2)
			{
				gx = cx;
				gy = nh2;
				//debug writefln("top middle");
			}
			// left
			if (cx > nw2)
			{
				gx = nw2;
				gy = nh2;
				//debug writefln("top left");
			}
			// right
			if (cx < x)
			{
				gx = x;
				gy = nh2;
				//debug writefln("top right");
			}
		}
		// box frame below radius frame
		else if (y2 > cy)
		{
			// middle
			if (cx >= x && cx <= nw2)
			{
				gx = cx;
				gy = y;
				//debug writefln("bottom middle");
			}
			// left
			if (cx > nw2)
			{
				gx = nw2;
				gy = y; 
				//debug writefln("bottom left");
			}
			// right
			if (cx < x)
			{
				gx = x;
				gy = y;
				//debug writefln("bottom right");
			}
		}
		
		// boxframe is on right middle of radframe
		if (cx < x2 && nh2 > cy && y < cy)
		{
			gx = x;
			gy = cy;
			//debug writefln("right middle");
		}
		// boxframe is on left middle of radframe
		else if (cx > x2 && nh2 > cy && y < cy)
		{
			gx = nw2;
			gy = cy;
			//debug writefln("left middle");
		}
	
		if (arc.phy.phyutil.distance(cx, cy, gx, gy) <= radius)
			return true; 

		return false;
}

/// determine whether x and y are within given box
bool boxXYCollision(float argX, float argY, 
					float x, float y, float width, float height)
{
	if (argX > x && argX < x + width)
		if (argY > y && argY < y + height)
			return true;

    return false;
}

// CIRCLES ///////////////////////////////////////////////////////////

/// determines whether or not 2 circles have collided
bool circleCircleCollision(	float cx1, float cy1, float rad1,
							float cx2, float cy2, float rad2)
{
	if (arc.phy.phyutil.distance(cx1, cy1, cx2, cy2) <= (rad1 + rad2))
		return true; 
	
	return false;  
}

/// determine whether point x, y is within circle
bool circleXYCollision(float x, float y, float cx, float cy, float rad)
{
	if (arc.phy.phyutil.distance(cx, cy, x, y) <= rad)
		return true;

	return false;
}

// LINE ///////////////////////////////////////////////////////////


// Line Line Collision /////////////////////////////////
// http://www.geometryalgorithms.com/Archive/algorithm_0104/algorithm_0104B.htm#intersect2D_SegSeg()
// intersect2D_2Segments(): the intersection of 2 finite 2D segments
//    Input:  two finite segments S1 and S2
//    Output: *I0 = intersect point (when it exists)
//    Return: 0=disjoint (no intersect)
//            1=intersect in unique point I0
//            2=overlap in segment from I0 to I1
int lineLineCollision(Point S1P0, Point S1P1, Point S2P0, Point S2P1, inout Point I0)
{
    Point    u = S1P1 - S1P0;
    Point    v = S2P1 - S2P0;
    Point    w = S1P0 - S2P0;
    float     D = u.perpProduct(v);

    // test if they are parallel (includes either being a point)
    if (fabs(D) < EPSILON) {          // S1 and S2 are parallel
        if (u.perpProduct(w) != 0 || v.perpProduct(w) != 0) {
            return false;                   // they are NOT collinear
        }
        // they are collinear or degenerate
        // check if they are degenerate points
        float du = u.dotProduct(u);
        float dv = v.dotProduct(v);
		
        if (du==0 && dv==0) {           // both segments are points
            if (S1P0 != S2P0)         // they are distinct points
                return false;
            I0 = S1P0;                // they are the same point
            return true;
        }
        if (du==0) {                    // S1 is a single point
            if (inSegment(S1P0, S2P0, S2P1) == 0)  // but is not in S2
                return false;
            I0 = S1P0;
            return true;
        }
        if (dv==0) {                    // S2 a single point
            if (inSegment(S2P0, S1P0, S1P1) == 0)  // but is not in S1
                return false;
            I0 = S2P0;
            return true;
        }
		
        // they are collinear segments - get overlap (or not)
        float t0, t1;                   // endpoints of S1 in eqn for S2
        Point w2 = S1P1 - S2P0;
        if (v.x != 0) {
                t0 = w.x / v.x;
                t1 = w2.x / v.x;
        }
        else {
                t0 = w.y / v.y;
                t1 = w2.y / v.y;
        }
        if (t0 > t1) {                  // must have t0 smaller than t1
                float t=t0; t0=t1; t1=t;    // swap if not
        }
        if (t0 > 1 || t1 < 0) {
            return 0;     // NO overlap
        }
        t0 = t0<0? 0 : t0;              // clip to min 0
        t1 = t1>1? 1 : t1;              // clip to max 1
        if (t0 == t1) {                 // intersect is a point
			Point tmp = t0 * v;
            I0 = S2P0 + tmp;
            return true;
        }

        // they overlap in a valid subsegment
		Point tmp1 = t0 * v;
		Point tmp2 = t1 * v; 
		
        I0 = S2P0 + tmp1;
        //I1 = S2P0 + tmp2;
        return true;
    }

    // the segments are skew and may intersect in a point
    // get the intersect parameter for S1
    float     sI = v.perpProduct(w) / D;
    if (sI < 0 || sI > 1)               // no intersect with S1
        return 0;

    // get the intersect parameter for S2
    float     tI = u.perpProduct(w) / D;
    if (tI < 0 || tI > 1)               // no intersect with S2
        return 0;

	Point tmp = sI * u; 
    I0 = S1P0 + tmp;               // compute S1 intersect point
    return 1;
}
//===================================================================

// inSegment(): determine if a point is inside a segment
//    Input:  a point P, and a collinear segment S
//    Return: true = P is inside S
//            false = P is not inside S
bool inSegment( inout Point P, inout Point S0, inout Point S1)
{
    if (S0.x != S1.x) {    // S is not vertical
        if (S0.x <= P.x && P.x <= S1.x)
            return true;
        if (S0.x >= P.x && P.x >= S1.x)
            return true;
    }
    else {    // S is vertical, so test y coordinate
        if (S0.y <= P.y && P.y <= S1.y)
            return true;
        if (S0.y >= P.y && P.y >= S1.y)
            return true;
    }
    return false;
}



// POINT //////////////////////////////////////////////////////////

/// returns true if this point is inside of this polygon
bool polygonPointCollision(inout Point p, Point[] points)
{
	bool c = false;
	int i, j;
		
	for (i = 0, j = points.length-1; i < points.length; j = i++) 
	{
		if ((((points[i].y<=p.y) && (p.y < points[j].y)) ||
			((points[j].y<=p.y) && (p.y < points[i].y))) &&
			(p.x < (points[j].x - points[i].x) * (p.y - points[i].y) / (points[j].y - points[i].y) + points[i].x))
		{
		  c = !c;
		}
	}
		
	return c;
}


