/*
normally your program is a while loop polling for SDL event and drawing + some game logic, so the 
while loop will run as fast as the machine can, and you'll get a high frame-rate and high CPU usage, 
the solution is to limit the frame-rate by putting an SDL_Delay in the event loop, the time slept by 
SDL_Delay should be variable(i.e. dont use SDL_Delay(30)).

for example, if you want to achieve a framerate of 30, then the time spent in each frame is 
1000/30 ~= 33 ms. if the time spent in drawing + game_logic = 5, then you should sleep 33 - 5 = 28 ms,
 and so on.
here are the steps:
1- calculate frametime from the desired frame rate(frametime = (1000/fps) ms).
2- calculate time spent each frame.
3- sleep alittle so that the time spent in the whole frame = desired frame time(SDL_Delay(frametime - time_spent_this_frame)).

I hope my explaination was clear.
you can do it yourself as I explained, or you can use a library such as sdl_gfx (I know it seems odd, but sdl_gfx contains a frame-limiting module).

*/

/*******************************************************************************

Description: Code relating to time

Authors: Clay Smith

Liscense: See <a href="http://www.opensource.org/licenses/zlib-license.php">zlib/libpng license</a>

Examples:
--------------------
import arc.phy.time; 

int main() {

	Blinker blink = new Blinker;
	Time time = new Time; 

	while (!done)
	{
		blink.process(5);
		time.process();

		// elapsed time since last frame in seconds
		time.elapsedSec();

		// elapsed time since last frame in milliseconds
		time.elapsedMilli();

		// will slow down frame until delay is met
		time.delay(10);

		// blink is on every 5 seconds
		if (blink.on)
		{
			// code here is called every 5 sec
		}
	}

   return 0;
}
--------------------

*******************************************************************************/


module arc.phy.time; 

private import derelict.sdl.sdl; 
debug import std.stdio;

/// Code good for anything that can blink
class Blinker 
{
  public:
   bool   on = false;
   double lastTime = 0, currTime = 0;
   double totalsec = 0.0f; 
   
   this()
   {
      lastTime = SDL_GetTicks();
      currTime = SDL_GetTicks()+.01; // make sure current starts out bigger than lastTime
   }

	/// blinker is on every # of seconds
   void process(float argSeconds)
   {
      lastTime = currTime; // last time equals what curr time was
      currTime = SDL_GetTicks(); // update curr time to the current time
   
      float seconds = currTime - lastTime;
      seconds /= 1000;
      
      totalsec += seconds; 
      
      on = false; 
      
      if(totalsec > argSeconds ) // if totalsec has elapsed since the last time
      {
         on = true; 
         totalsec = 0; 
      }
   }
}


/// Gives elapsed time, gives frames per second and allows control over frames per seconds
class Time
{
  private:
	double totalSec = 0;
	double prevTime, currTime;
	uint frames = 0;
	uint fps = 60;
	uint targetFPS = 60; 

  public:
	/// initialize frame per second amount
	this()
	{
		prevTime = SDL_GetTicks()-.1;
		currTime = SDL_GetTicks();
	}

	/// call once per frame for time to work
	void process()
	{
		// calculate our current frames per second
		prevTime = currTime;
		currTime = SDL_GetTicks();

		frames++;

		totalSec += (currTime - prevTime)/1000;

		if (totalSec >= 1)
		{
			fps = frames;
			frames = 0;
			totalSec = 0;
		}
		
	}

	/// force every frame to wait a certain amount of milliseconds before continuing
	void delay(uint argMilli)
	{			
		// force wait if we did not wait 'argMilli' amount of time
		if ( (argMilli - (currTime - prevTime)) > 0)
		{
			SDL_Delay(cast(uint)((argMilli - (currTime - prevTime))));
			prevTime = currTime; 
			currTime = SDL_GetTicks();
		}
	}

	/// return current frame rate
	uint getFPS() { return fps; }

	/// elapsed time in seconds since last frame
	uint getElapsedSec() { return (cast(uint)((currTime - prevTime)/1000)); }

	/// elapsed time in milliseconds since last frame
	uint getElapsedMil() { return (cast(uint)(currTime - prevTime)); }

  private:

}


/// Returns the number of milliseconds since the creation of the application
uint getTime()
{
	return SDL_GetTicks(); 
}