/*******************************************************************************

Description: A triangle class that allows you to do trigonomitry and collision 

Author: Clay Smith

License: zlib/png

*******************************************************************************/

module arc.phy.triangle; 

private import std.math; 

private import 	arc.phy.point,
				arc.phy.phyutil; // for distance

private import derelict.opengl.gl; 

private import arc.io.log;  

enum
{
	SIDEA=0,
	SIDEB,
	SIDEC
}

enum
{
	ANGLEA=0,
	ANGLEB,
	ANGLEC
}

enum 
{
	A = 0, 
	B, 
	C
}

/*******************************************************************************

   The triangle is made up of 3 points

********************************************************************************/
struct Triangle
{
  private:
	// outline before rotation, zoom, position, and pivot 
	Point[] origPoints; 

	// points after rotation, zoom and position
	Point[] points; 

  public:
	static Triangle opCall(Point a, Point b, Point c)
	{
		Triangle t;
		t.setPoints(a,b,c); 
		return t;
	}

	// rotate points 
	void transformPoints(float angle, Point zoom, Point pivot, Point position)
	{
		for(int i = 0; i < 3; i++)
		{
			points[i] = origPoints[i].getRotation(pivot, angle) * zoom + position;
		} 
	}

   /// draw -- triangle
   void draw(ushort r, ushort g, ushort b, ushort a, bool fill)
   {
		glDisable(GL_TEXTURE_2D); 
		glColor4ub(r,g,b,a);

		if (fill)
		{
			// Draw the triangle
			glBegin(GL_POLYGON);
			 glVertex2f(points[A].getX, points[A].getY);
			 glVertex2f(points[B].getX, points[B].getY);
			 glVertex2f(points[C].getX, points[C].getY);
			glEnd();   
		}
		else
		{
			// Draw the triangle
			glBegin(GL_LINE_LOOP);
			 glVertex2f(points[A].getX, points[A].getY);
			 glVertex2f(points[B].getX, points[B].getY);
			 glVertex2f(points[C].getX, points[C].getY);
			glEnd();     
		}

		glEnable(GL_TEXTURE_2D);
   }

   /// get -- points
   Point getPointA() { return points[A]; }
   Point getPointB() { return points[B]; }
   Point getPointC() { return points[C]; }

   float getPointAx() { return points[A].getX; }
   float getPointAy() { return points[A].getY; }
   float getPointBx() { return points[B].getX; }
   float getPointBy() { return points[B].getY; }
   float getPointCx() { return points[C].getX; }
   float getPointCy() { return points[C].getY; }

   Point getPoint(int i)
   {
	   return points[i];
   }

   Point[] getPoints()
   {
	   return points;
   }

   /// set -- points
   void setPoints(Point vA, Point vB, Point vC)
   {
	  origPoints.length=3;
	  points.length=3;
	  
      points[A] = origPoints[A] = vA;
      points[B] = origPoints[B] = vB;
      points[C] = origPoints[C] = vC; 
   }

   /// get -- solves the lengths given three points
   float[] getLengths()
   in
   {
      // assert that our data is valid
      assert(!points[A].isNaN());
      assert(!points[B].isNaN());
      assert(!points[C].isNaN()); 
   }
   body
   {
		float[] v; v.length=3;
		v[SIDEA] = points[B].distance(points[C]);
		v[SIDEB] = points[A].distance(points[C]); 
		v[SIDEC] = points[A].distance(points[B]);
		return v; 
   }

   /// get -- solves the angles of a triangle given the lengths
   float[] getAngles(float[] s)   
   in
   {
	  assert(s.length == 3); 
      assert(s[SIDEA] != float.nan);
      assert(s[SIDEB] != float.nan);   
      assert(s[SIDEC] != float.nan); 
   }
   body
   {
		float v[]; v.length = 3; 
		v[ANGLEA] = acos( (pow(s[SIDEA],2.0) - pow(s[SIDEB],2.0) - pow(s[SIDEC],2.0)) / ( -2 * s[SIDEB] * s[SIDEC]) ) / DEGREE_TO_RADIAN;
		v[ANGLEB] = acos( (pow(s[SIDEB],2.0) - pow(s[SIDEA],2.0) - pow(s[SIDEC],2.0)) / ( -2 * s[SIDEA] * s[SIDEC]) ) / DEGREE_TO_RADIAN;
		v[ANGLEC] = acos( (pow(s[SIDEC],2.0) - pow(s[SIDEA],2.0) - pow(s[SIDEB],2.0)) / ( -2 * s[SIDEA] * s[SIDEB]) ) / DEGREE_TO_RADIAN;
		return v;
   }

}


unittest
{
   debug writefln("Triangle: Unit Test Begin...");

   Point A, B, C; 

   A = Point(40, 50);
   B = Point(51, 90); 
   C = Point(50, 45); 
   
   Triangle tri; 

   tri.setPoints(A, B, C); 

   float[] l = tri.getLengths(); 
   float[] a = tri.getAngles(l); 

   debug writefln("Triangle: Lengths are: a = ", l[SIDEA], " b = ", l[SIDEB], " c = ", l[SIDEC]);
   debug writefln("Triangle: Angles are: A = ", a[ANGLEA], " B = ", a[ANGLEC], " C = ", a[ANGLEB]);

   debug writefln("Triangle: Unit Test End...");  
}
